from __future__ import absolute_import, unicode_literals

import json

from mopidy.models import immutable


class ModelJSONEncoder(json.JSONEncoder):

    """
    Automatically serialize Mopidy models to JSON.

    Usage::

        >>> import json
        >>> json.dumps({'a_track': Track(name='name')}, cls=ModelJSONEncoder)
        '{"a_track": {"__model__": "Track", "name": "name"}}'

    """

    def default(self, obj):
        if isinstance(obj, immutable.ImmutableObject):
            return obj.serialize()
        return json.JSONEncoder.default(self, obj)


def model_json_decoder(dct):
    """
    Automatically deserialize Mopidy models from JSON.

    Usage::

        >>> import json
        >>> json.loads(
        ...     '{"a_track": {"__model__": "Track", "name": "name"}}',
        ...     object_hook=model_json_decoder)
        {u'a_track': Track(artists=[], name=u'name')}

    """
    if '__model__' in dct:
        model_name = dct.pop('__model__')
        if model_name in immutable._models:
            cls = immutable._models[model_name]
            return cls(**dct)
    return dct



from __future__ import absolute_import, unicode_literals

import logging
import string

logger = logging.getLogger(__name__)

try:
    import dbus
except ImportError:
    dbus = None

_AVAHI_IF_UNSPEC = -1
_AVAHI_PROTO_UNSPEC = -1
_AVAHI_PUBLISHFLAGS_NONE = 0


def _is_loopback_address(host):
    return (
        host.startswith('127.') or
        host.startswith('::ffff:127.') or
        host == '::1')


def _convert_text_list_to_dbus_format(text_list):
    array = dbus.Array(signature='ay')
    for text in text_list:
        array.append([dbus.Byte(ord(c)) for c in text])
    return array


class Zeroconf(object):

    """Publish a network service with Zeroconf.

    Currently, this only works on Linux using Avahi via D-Bus.

    :param str name: human readable name of the service, e.g. 'MPD on neptune'
    :param str stype: service type, e.g. '_mpd._tcp'
    :param int port: TCP port of the service, e.g. 6600
    :param str domain: local network domain name, defaults to ''
    :param str host: interface to advertise the service on, defaults to ''
    :param text: extra information depending on ``stype``, defaults to empty
        list
    :type text: list of str
    """

    def __init__(self, name, stype, port, domain='', host='', text=None):
        self.stype = stype
        self.port = port
        self.domain = domain
        self.host = host
        self.text = text or []

        self.bus = None
        self.server = None
        self.group = None
        self.display_hostname = None
        self.name = None

        if dbus:
            try:
                self.bus = dbus.SystemBus()
                self.server = dbus.Interface(
                    self.bus.get_object('org.freedesktop.Avahi', '/'),
                    'org.freedesktop.Avahi.Server')
                self.display_hostname = '%s' % self.server.GetHostName()
                self.name = string.Template(name).safe_substitute(
                    hostname=self.display_hostname, port=port)
            except dbus.exceptions.DBusException as e:
                logger.debug('%s: Server failed: %s', self, e)

    def __str__(self):
        return 'Zeroconf service "%s" (%s at [%s]:%d)' % (
            self.name, self.stype, self.host, self.port)

    def publish(self):
        """Publish the service.

        Call when your service starts.
        """

        if _is_loopback_address(self.host):
            logger.debug(
                '%s: Publish on loopback interface is not supported.', self)
            return False

        if not dbus:
            logger.debug('%s: dbus not installed; publish failed.', self)
            return False

        if not self.bus:
            logger.debug('%s: Bus not available; publish failed.', self)
            return False

        if not self.server:
            logger.debug('%s: Server not available; publish failed.', self)
            return False

        try:
            if not self.bus.name_has_owner('org.freedesktop.Avahi'):
                logger.debug(
                    '%s: Avahi service not running; publish failed.', self)
                return False

            self.group = dbus.Interface(
                self.bus.get_object(
                    'org.freedesktop.Avahi', self.server.EntryGroupNew()),
                'org.freedesktop.Avahi.EntryGroup')

            self.group.AddService(
                _AVAHI_IF_UNSPEC, _AVAHI_PROTO_UNSPEC,
                dbus.UInt32(_AVAHI_PUBLISHFLAGS_NONE),
                self.name, self.stype,
                self.domain, self.host, dbus.UInt16(self.port),
                _convert_text_list_to_dbus_format(self.text))

            self.group.Commit()
            logger.debug('%s: Published', self)
            return True
        except dbus.exceptions.DBusException as e:
            logger.debug('%s: Publish failed: %s', self, e)
            return False

    def unpublish(self):
        """Unpublish the service.

        Call when your service shuts down.
        """

        if self.group:
            try:
                self.group.Reset()
                logger.debug('%s: Unpublished', self)
            except dbus.exceptions.DBusException as e:
                logger.debug('%s: Unpublish failed: %s', self, e)
            finally:
                self.group = None



from __future__ import absolute_import, unicode_literals

import unittest

from mopidy import compat
from mopidy.core import HistoryController
from mopidy.internal.models import HistoryState, HistoryTrack
from mopidy.models import Artist, Ref, Track


class PlaybackHistoryTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.tracks = [
            Track(uri='dummy1:a', name='foo',
                  artists=[Artist(name='foober'), Artist(name='barber')]),
            Track(uri='dummy2:a', name='foo'),
            Track(uri='dummy3:a', name='bar')
        ]
        self.history = HistoryController()

    def test_add_track(self):
        self.history._add_track(self.tracks[0])
        self.assertEqual(self.history.get_length(), 1)

        self.history._add_track(self.tracks[1])
        self.assertEqual(self.history.get_length(), 2)

        self.history._add_track(self.tracks[2])
        self.assertEqual(self.history.get_length(), 3)

    def test_non_tracks_are_rejected(self):
        with self.assertRaises(TypeError):
            self.history._add_track(object())

        self.assertEqual(self.history.get_length(), 0)

    def test_history_entry_contents(self):
        track = self.tracks[0]
        self.history._add_track(track)

        result = self.history.get_history()
        (timestamp, ref) = result[0]

        self.assertIsInstance(timestamp, compat.integer_types)
        self.assertEqual(track.uri, ref.uri)
        self.assertIn(track.name, ref.name)
        for artist in track.artists:
            self.assertIn(artist.name, ref.name)


class CoreHistorySaveLoadStateTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.tracks = [
            Track(uri='dummy1:a', name='foober'),
            Track(uri='dummy2:a', name='foo'),
            Track(uri='dummy3:a', name='bar')
        ]

        self.refs = []
        for t in self.tracks:
            self.refs.append(Ref.track(uri=t.uri, name=t.name))

        self.history = HistoryController()

    def test_save(self):
        self.history._add_track(self.tracks[2])
        self.history._add_track(self.tracks[1])

        value = self.history._save_state()

        self.assertEqual(len(value.history), 2)
        # last in, first out
        self.assertEqual(value.history[0].track, self.refs[1])
        self.assertEqual(value.history[1].track, self.refs[2])

    def test_load(self):
        state = HistoryState(history=[
            HistoryTrack(timestamp=34, track=self.refs[0]),
            HistoryTrack(timestamp=45, track=self.refs[2]),
            HistoryTrack(timestamp=56, track=self.refs[1])])
        coverage = ['history']
        self.history._load_state(state, coverage)

        hist = self.history.get_history()
        self.assertEqual(len(hist), 3)
        self.assertEqual(hist[0], (34, self.refs[0]))
        self.assertEqual(hist[1], (45, self.refs[2]))
        self.assertEqual(hist[2], (56, self.refs[1]))

        # after import, adding more tracks must be possible
        self.history._add_track(self.tracks[1])
        hist = self.history.get_history()
        self.assertEqual(len(hist), 4)
        self.assertEqual(hist[0][1], self.refs[1])
        self.assertEqual(hist[1], (34, self.refs[0]))
        self.assertEqual(hist[2], (45, self.refs[2]))
        self.assertEqual(hist[3], (56, self.refs[1]))

    def test_load_invalid_type(self):
        with self.assertRaises(TypeError):
            self.history._load_state(11, None)

    def test_load_none(self):
        self.history._load_state(None, None)



from __future__ import absolute_import, unicode_literals

from tests.mpd import protocol


class MountTest(protocol.BaseTestCase):

    def test_mount(self):
        self.send_request('mount my_disk /dev/sda')
        self.assertEqualResponse('ACK [0@0] {mount} Not implemented')

    def test_unmount(self):
        self.send_request('unmount my_disk')
        self.assertEqualResponse('ACK [0@0] {unmount} Not implemented')

    def test_listmounts(self):
        self.send_request('listmounts')
        self.assertEqualResponse('ACK [0@0] {listmounts} Not implemented')

    def test_listneighbors(self):
        self.send_request('listneighbors')
        self.assertEqualResponse('ACK [0@0] {listneighbors} Not implemented')



from __future__ import absolute_import, unicode_literals

from tests.mpd import protocol


class StickersHandlerTest(protocol.BaseTestCase):

    def test_sticker_get(self):
        self.send_request(
            'sticker get "song" "file:///dev/urandom" "a_name"')
        self.assertEqualResponse('ACK [0@0] {sticker} Not implemented')

    def test_sticker_set(self):
        self.send_request(
            'sticker set "song" "file:///dev/urandom" "a_name" "a_value"')
        self.assertEqualResponse('ACK [0@0] {sticker} Not implemented')

    def test_sticker_delete_with_name(self):
        self.send_request(
            'sticker delete "song" "file:///dev/urandom" "a_name"')
        self.assertEqualResponse('ACK [0@0] {sticker} Not implemented')

    def test_sticker_delete_without_name(self):
        self.send_request(
            'sticker delete "song" "file:///dev/urandom"')
        self.assertEqualResponse('ACK [0@0] {sticker} Not implemented')

    def test_sticker_list(self):
        self.send_request(
            'sticker list "song" "file:///dev/urandom"')
        self.assertEqualResponse('ACK [0@0] {sticker} Not implemented')

    def test_sticker_find(self):
        self.send_request(
            'sticker find "song" "file:///dev/urandom" "a_name"')
        self.assertEqualResponse('ACK [0@0] {sticker} Not implemented')



from __future__ import absolute_import, unicode_literals

import os

from mopidy import compat


def path_to_data_dir(name):
    if not isinstance(name, bytes):
        name = name.encode('utf-8')
    path = os.path.dirname(__file__)
    path = os.path.join(path, b'data')
    path = os.path.abspath(path)
    return os.path.join(path, name)


class IsA(object):

    def __init__(self, klass):
        self.klass = klass

    def __eq__(self, rhs):
        try:
            return isinstance(rhs, self.klass)
        except TypeError:
            return type(rhs) == type(self.klass)  # noqa

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def __repr__(self):
        return str(self.klass)


any_int = IsA(compat.integer_types)
any_str = IsA(compat.string_types)
any_unicode = IsA(compat.text_type)



# encoding: utf-8

"""Mopidy documentation build configuration file"""

from __future__ import absolute_import, unicode_literals

import os
import sys


# -- Workarounds to have autodoc generate API docs ----------------------------

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__) + '/../'))


class Mock(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    def __or__(self, other):
        return Mock()

    @classmethod
    def __getattr__(self, name):
        if name == 'get_system_config_dirs':  # GLib.get_system_config_dirs()
            return list
        elif name == 'get_user_config_dir':  # GLib.get_user_config_dir()
            return str
        else:
            return Mock()


MOCK_MODULES = [
    'dbus',
    'dbus.mainloop',
    'dbus.mainloop.glib',
    'dbus.service',
    'mopidy.internal.gi',
]
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()


# -- Custom Sphinx object types -----------------------------------------------

def setup(app):
    from sphinx.ext.autodoc import cut_lines
    app.connect(b'autodoc-process-docstring', cut_lines(4, what=['module']))
    app.add_object_type(
        b'confval', 'confval',
        objname='configuration value',
        indextemplate='pair: %s; configuration value')


# -- General configuration ----------------------------------------------------

needs_sphinx = '1.3'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.extlinks',
    'sphinx.ext.graphviz',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'

project = 'Mopidy'
copyright = '2009-2019, Stein Magnus Jodal and contributors'

from mopidy.internal.versioning import get_version
release = get_version()
version = '.'.join(release.split('.')[:2])

# To make the build reproducible, avoid using today's date in the manpages
today = '2019'

exclude_trees = ['_build']

pygments_style = 'sphinx'

modindex_common_prefix = ['mopidy.']


# -- Options for HTML output --------------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

html_use_modindex = True
html_use_index = True
html_split_index = False
html_show_sourcelink = True

htmlhelp_basename = 'Mopidy'


# -- Options for LaTeX output -------------------------------------------------

latex_documents = [
    (
        'index',
        'Mopidy.tex',
        'Mopidy Documentation',
        'Stein Magnus Jodal and contributors',
        'manual'
    ),
]


# -- Options for manpages output ----------------------------------------------

man_pages = [
    (
        'command',
        'mopidy',
        'music server',
        '',
        '1'
    ),
]


# -- Options for extlink extension --------------------------------------------

extlinks = {
    'issue': ('https://github.com/mopidy/mopidy/issues/%s', '#'),
    'commit': ('https://github.com/mopidy/mopidy/commit/%s', 'commit '),
    'js': ('https://github.com/mopidy/mopidy.js/issues/%s', 'mopidy.js#'),
    'mpris': (
        'https://github.com/mopidy/mopidy-mpris/issues/%s', 'mopidy-mpris#'),
    'discuss': (
        'https://discourse.mopidy.com/t/%s', 'discourse.mopidy.com/t/'),
}


# -- Options for intersphinx extension ----------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/2', None),
    'pykka': ('https://www.pykka.org/en/latest/', None),
    'tornado': ('http://www.tornadoweb.org/en/stable/', None),
}

# -- Options for linkcheck builder -------------------------------------------

linkcheck_ignore = [  # Some sites work in browser but linkcheck fails.
    r'http://localhost:\d+/',
    r'http://wiki.commonjs.org',
    r'http://vk.com',
    r'http://$']

linkcheck_anchors = False  # This breaks on links that use # for other stuff



from __future__ import unicode_literals

import logging
import os
import sys
import urllib2

from mopidy import backend, exceptions, models
from mopidy.audio import scan, tags
from mopidy.internal import path


logger = logging.getLogger(__name__)
FS_ENCODING = sys.getfilesystemencoding()


class FileLibraryProvider(backend.LibraryProvider):
    """Library for browsing local files."""

    # TODO: get_images that can pull from metadata and/or .folder.png etc?
    # TODO: handle playlists?

    @property
    def root_directory(self):
        if not self._media_dirs:
            return None
        elif len(self._media_dirs) == 1:
            uri = path.path_to_uri(self._media_dirs[0]['path'])
        else:
            uri = 'file:root'
        return models.Ref.directory(name='Files', uri=uri)

    def __init__(self, backend, config):
        super(FileLibraryProvider, self).__init__(backend)
        self._media_dirs = list(self._get_media_dirs(config))
        self._show_dotfiles = config['file']['show_dotfiles']
        self._excluded_file_extensions = tuple(
            bytes(file_ext.lower())
            for file_ext in config['file']['excluded_file_extensions'])
        self._follow_symlinks = config['file']['follow_symlinks']

        self._scanner = scan.Scanner(
            timeout=config['file']['metadata_timeout'])

    def browse(self, uri):
        logger.debug('Browsing files at: %s', uri)
        result = []
        local_path = path.uri_to_path(uri)

        if local_path == 'root':
            return list(self._get_media_dirs_refs())

        if not self._is_in_basedir(os.path.realpath(local_path)):
            logger.warning(
                'Rejected attempt to browse path (%s) outside dirs defined '
                'in file/media_dirs config.', uri)
            return []

        for dir_entry in os.listdir(local_path):
            child_path = os.path.join(local_path, dir_entry)
            uri = path.path_to_uri(child_path)

            if not self._show_dotfiles and dir_entry.startswith(b'.'):
                continue

            if (self._excluded_file_extensions and
                    dir_entry.endswith(self._excluded_file_extensions)):
                continue

            if os.path.islink(child_path) and not self._follow_symlinks:
                logger.debug('Ignoring symlink: %s', uri)
                continue

            if not self._is_in_basedir(os.path.realpath(child_path)):
                logger.debug('Ignoring symlink to outside base dir: %s', uri)
                continue

            name = dir_entry.decode(FS_ENCODING, 'replace')
            if os.path.isdir(child_path):
                result.append(models.Ref.directory(name=name, uri=uri))
            elif os.path.isfile(child_path):
                result.append(models.Ref.track(name=name, uri=uri))

        def order(item):
            return (item.type != models.Ref.DIRECTORY, item.name)
        result.sort(key=order)

        return result

    def lookup(self, uri):
        logger.debug('Looking up file URI: %s', uri)
        local_path = path.uri_to_path(uri)

        try:
            result = self._scanner.scan(uri)
            track = tags.convert_tags_to_track(result.tags).copy(
                uri=uri, length=result.duration)
        except exceptions.ScannerError as e:
            logger.warning('Failed looking up %s: %s', uri, e)
            track = models.Track(uri=uri)

        if not track.name:
            filename = os.path.basename(local_path)
            name = urllib2.unquote(filename).decode(FS_ENCODING, 'replace')
            track = track.copy(name=name)

        return [track]

    def _get_media_dirs(self, config):
        for entry in config['file']['media_dirs']:
            media_dir = {}
            media_dir_split = entry.split('|', 1)
            local_path = path.expand_path(
                media_dir_split[0].encode(FS_ENCODING))

            if not local_path:
                logger.debug(
                    'Failed expanding path (%s) from file/media_dirs config '
                    'value.',
                    media_dir_split[0])
                continue
            elif not os.path.isdir(local_path):
                logger.warning(
                    '%s is not a directory. Please create the directory or '
                    'update the file/media_dirs config value.', local_path)
                continue

            media_dir['path'] = local_path
            if len(media_dir_split) == 2:
                media_dir['name'] = media_dir_split[1]
            else:
                # TODO Mpd client should accept / in dir name
                media_dir['name'] = media_dir_split[0].replace(os.sep, '+')

            yield media_dir

    def _get_media_dirs_refs(self):
        for media_dir in self._media_dirs:
            yield models.Ref.directory(
                name=media_dir['name'],
                uri=path.path_to_uri(media_dir['path']))

    def _is_in_basedir(self, local_path):
        return any(
            path.is_path_inside_base_dir(local_path, media_dir['path'])
            for media_dir in self._media_dirs)



"""A dummy backend for use in tests.

This backend implements the backend API in the simplest way possible.  It is
used in tests of the frontends.
"""

from __future__ import absolute_import, unicode_literals

import pykka

from mopidy import backend
from mopidy.models import Playlist, Ref, SearchResult


def create_proxy(config=None, audio=None):
    return DummyBackend.start(config=config, audio=audio).proxy()


class DummyBackend(pykka.ThreadingActor, backend.Backend):

    def __init__(self, config, audio):
        super(DummyBackend, self).__init__()

        self.library = DummyLibraryProvider(backend=self)
        if audio:
            self.playback = backend.PlaybackProvider(audio=audio, backend=self)
        else:
            self.playback = DummyPlaybackProvider(audio=audio, backend=self)
        self.playlists = DummyPlaylistsProvider(backend=self)

        self.uri_schemes = ['dummy']


class DummyLibraryProvider(backend.LibraryProvider):
    root_directory = Ref.directory(uri='dummy:/', name='dummy')

    def __init__(self, *args, **kwargs):
        super(DummyLibraryProvider, self).__init__(*args, **kwargs)
        self.dummy_library = []
        self.dummy_get_distinct_result = {}
        self.dummy_browse_result = {}
        self.dummy_find_exact_result = SearchResult()
        self.dummy_search_result = SearchResult()

    def browse(self, path):
        return self.dummy_browse_result.get(path, [])

    def get_distinct(self, field, query=None):
        return self.dummy_get_distinct_result.get(field, set())

    def lookup(self, uri):
        return [t for t in self.dummy_library if uri == t.uri]

    def refresh(self, uri=None):
        pass

    def search(self, query=None, uris=None, exact=False):
        if exact:  # TODO: remove uses of dummy_find_exact_result
            return self.dummy_find_exact_result
        return self.dummy_search_result


class DummyPlaybackProvider(backend.PlaybackProvider):

    def __init__(self, *args, **kwargs):
        super(DummyPlaybackProvider, self).__init__(*args, **kwargs)
        self._uri = None
        self._time_position = 0

    def pause(self):
        return True

    def play(self):
        return self._uri and self._uri != 'dummy:error'

    def change_track(self, track):
        """Pass a track with URI 'dummy:error' to force failure"""
        self._uri = track.uri
        self._time_position = 0
        return True

    def prepare_change(self):
        pass

    def resume(self):
        return True

    def seek(self, time_position):
        self._time_position = time_position
        return True

    def stop(self):
        self._uri = None
        return True

    def get_time_position(self):
        return self._time_position


class DummyPlaylistsProvider(backend.PlaylistsProvider):

    def __init__(self, backend):
        super(DummyPlaylistsProvider, self).__init__(backend)
        self._playlists = []
        self._allow_save = True

    def set_dummy_playlists(self, playlists):
        """For tests using the dummy provider through an actor proxy."""
        self._playlists = playlists

    def set_allow_save(self, enabled):
        self._allow_save = enabled

    def as_list(self):
        return [
            Ref.playlist(uri=pl.uri, name=pl.name) for pl in self._playlists]

    def get_items(self, uri):
        playlist = self.lookup(uri)
        if playlist is None:
            return
        return [
            Ref.track(uri=t.uri, name=t.name) for t in playlist.tracks]

    def lookup(self, uri):
        for playlist in self._playlists:
            if playlist.uri == uri:
                return playlist

    def refresh(self):
        pass

    def create(self, name):
        playlist = Playlist(name=name, uri='dummy:%s' % name)
        self._playlists.append(playlist)
        return playlist

    def delete(self, uri):
        playlist = self.lookup(uri)
        if playlist:
            self._playlists.remove(playlist)

    def save(self, playlist):
        if not self._allow_save:
            return None

        old_playlist = self.lookup(playlist.uri)

        if old_playlist is not None:
            index = self._playlists.index(old_playlist)
            self._playlists[index] = playlist
        else:
            self._playlists.append(playlist)

        return playlist



from __future__ import absolute_import, unicode_literals

import os
import shutil
import tempfile
import unittest

import mock

import pykka

import mopidy
from mopidy.core import Core
from mopidy.internal import models, storage, versioning
from mopidy.models import Track

from tests import dummy_mixer


class CoreActorTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.backend1 = mock.Mock()
        self.backend1.uri_schemes.get.return_value = ['dummy1']
        self.backend1.actor_ref.actor_class.__name__ = b'B1'

        self.backend2 = mock.Mock()
        self.backend2.uri_schemes.get.return_value = ['dummy2']
        self.backend2.actor_ref.actor_class.__name__ = b'B2'

        self.core = Core(mixer=None, backends=[self.backend1, self.backend2])

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def test_uri_schemes_has_uris_from_all_backends(self):
        result = self.core.get_uri_schemes()

        self.assertIn('dummy1', result)
        self.assertIn('dummy2', result)

    def test_backends_with_colliding_uri_schemes_fails(self):
        self.backend2.uri_schemes.get.return_value = ['dummy1', 'dummy2']

        self.assertRaisesRegexp(
            AssertionError,
            'Cannot add URI scheme "dummy1" for B2, '
            'it is already handled by B1',
            Core, mixer=None, backends=[self.backend1, self.backend2])

    def test_version(self):
        self.assertEqual(self.core.get_version(), versioning.get_version())


class CoreActorSaveLoadStateTest(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.state_file = os.path.join(self.temp_dir,
                                       b'core', b'state.json.gz')

        config = {
            'core': {
                'max_tracklist_length': 10000,
                'restore_state': True,
                'data_dir': self.temp_dir,
            }
        }

        os.mkdir(os.path.join(self.temp_dir, b'core'))

        self.mixer = dummy_mixer.create_proxy()
        self.core = Core(
            config=config, mixer=self.mixer, backends=[])

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()
        shutil.rmtree(self.temp_dir)

    def test_save_state(self):
        self.core.teardown()

        assert os.path.isfile(self.state_file)
        reload_data = storage.load(self.state_file)
        data = {}
        data['version'] = mopidy.__version__
        data['state'] = models.CoreState(
            tracklist=models.TracklistState(
                repeat=False, random=False,
                consume=False, single=False,
                next_tlid=1),
            history=models.HistoryState(),
            playback=models.PlaybackState(state='stopped',
                                          time_position=0),
            mixer=models.MixerState())
        assert data == reload_data

    def test_load_state_no_file(self):
        self.core.setup()

        assert self.core.mixer.get_mute() is None
        assert self.core.mixer.get_volume() is None
        assert self.core.tracklist._next_tlid == 1
        assert self.core.tracklist.get_repeat() is False
        assert self.core.tracklist.get_random() is False
        assert self.core.tracklist.get_consume() is False
        assert self.core.tracklist.get_single() is False
        assert self.core.tracklist.get_length() == 0
        assert self.core.playback._start_paused is False
        assert self.core.playback._start_at_position is None
        assert self.core.history.get_length() == 0

    def test_load_state_with_data(self):
        data = {}
        data['version'] = mopidy.__version__
        data['state'] = models.CoreState(
            tracklist=models.TracklistState(
                repeat=True, random=True,
                consume=False, single=False,
                tl_tracks=[models.TlTrack(tlid=12, track=Track(uri='a:a'))],
                next_tlid=14),
            history=models.HistoryState(history=[
                models.HistoryTrack(
                    timestamp=12,
                    track=models.Ref.track(uri='a:a', name='a')),
                models.HistoryTrack(
                    timestamp=13,
                    track=models.Ref.track(uri='a:b', name='b'))]),
            playback=models.PlaybackState(tlid=12, state='paused',
                                          time_position=432),
            mixer=models.MixerState(mute=True, volume=12))
        storage.dump(self.state_file, data)

        self.core.setup()

        assert self.core.mixer.get_mute() is True
        assert self.core.mixer.get_volume() == 12
        assert self.core.tracklist._next_tlid == 14
        assert self.core.tracklist.get_repeat() is True
        assert self.core.tracklist.get_random() is True
        assert self.core.tracklist.get_consume() is False
        assert self.core.tracklist.get_single() is False
        assert self.core.tracklist.get_length() == 1
        assert self.core.playback._start_paused is True
        assert self.core.playback._start_at_position == 432
        assert self.core.history.get_length() == 2

    def test_delete_state_file_on_restore(self):
        data = {}
        storage.dump(self.state_file, data)
        assert os.path.isfile(self.state_file)

        self.core.setup()

        assert not os.path.isfile(self.state_file)



from __future__ import absolute_import, unicode_literals

import unittest

from mopidy import exceptions


class ExceptionsTest(unittest.TestCase):

    def test_exception_can_include_message_string(self):
        exc = exceptions.MopidyException('foo')

        self.assertEqual(exc.message, 'foo')
        self.assertEqual(str(exc), 'foo')

    def test_backend_error_is_a_mopidy_exception(self):
        self.assert_(issubclass(
            exceptions.BackendError, exceptions.MopidyException))

    def test_extension_error_is_a_mopidy_exception(self):
        self.assert_(issubclass(
            exceptions.ExtensionError, exceptions.MopidyException))

    def test_find_error_is_a_mopidy_exception(self):
        self.assert_(issubclass(
            exceptions.FindError, exceptions.MopidyException))

    def test_find_error_can_store_an_errno(self):
        exc = exceptions.FindError('msg', errno=1234)

        self.assertEqual(exc.message, 'msg')
        self.assertEqual(exc.errno, 1234)

    def test_frontend_error_is_a_mopidy_exception(self):
        self.assert_(issubclass(
            exceptions.FrontendError, exceptions.MopidyException))

    def test_mixer_error_is_a_mopidy_exception(self):
        self.assert_(issubclass(
            exceptions.MixerError, exceptions.MopidyException))

    def test_scanner_error_is_a_mopidy_exception(self):
        self.assert_(issubclass(
            exceptions.ScannerError, exceptions.MopidyException))

    def test_audio_error_is_a_mopidy_exception(self):
        self.assert_(issubclass(
            exceptions.AudioException, exceptions.MopidyException))



from __future__ import unicode_literals


TAGTYPE_LIST = [
    'Artist',
    'ArtistSort',
    'Album',
    'AlbumArtist',
    'AlbumArtistSort',
    'Title',
    'Track',
    'Name',
    'Genre',
    'Date',
    'Composer',
    'Performer',
    'Disc',
    'MUSICBRAINZ_ARTISTID',
    'MUSICBRAINZ_ALBUMID',
    'MUSICBRAINZ_ALBUMARTISTID',
    'MUSICBRAINZ_TRACKID',
    'X-AlbumUri',
    'X-AlbumImage',
]



from __future__ import absolute_import, unicode_literals

import unittest

import mock

from mopidy import backend, core
from mopidy.models import Playlist, Ref, Track


class BasePlaylistsTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.plr1a = Ref.playlist(name='A', uri='dummy1:pl:a')
        self.plr1b = Ref.playlist(name='B', uri='dummy1:pl:b')
        self.plr2a = Ref.playlist(name='A', uri='dummy2:pl:a')
        self.plr2b = Ref.playlist(name='B', uri='dummy2:pl:b')

        self.pl1a = Playlist(name='A', tracks=[Track(uri='dummy1:t:a')])
        self.pl1b = Playlist(name='B', tracks=[Track(uri='dummy1:t:b')])
        self.pl2a = Playlist(name='A', tracks=[Track(uri='dummy2:t:a')])
        self.pl2b = Playlist(name='B', tracks=[Track(uri='dummy2:t:b')])

        self.sp1 = mock.Mock(spec=backend.PlaylistsProvider)
        self.sp1.as_list.return_value.get.return_value = [
            self.plr1a, self.plr1b]
        self.sp1.lookup.return_value.get.side_effect = [self.pl1a, self.pl1b]

        self.sp2 = mock.Mock(spec=backend.PlaylistsProvider)
        self.sp2.as_list.return_value.get.return_value = [
            self.plr2a, self.plr2b]
        self.sp2.lookup.return_value.get.side_effect = [self.pl2a, self.pl2b]

        self.backend1 = mock.Mock()
        self.backend1.actor_ref.actor_class.__name__ = 'Backend1'
        self.backend1.uri_schemes.get.return_value = ['dummy1']
        self.backend1.playlists = self.sp1

        self.backend2 = mock.Mock()
        self.backend2.actor_ref.actor_class.__name__ = 'Backend2'
        self.backend2.uri_schemes.get.return_value = ['dummy2']
        self.backend2.playlists = self.sp2

        # A backend without the optional playlists provider
        self.backend3 = mock.Mock()
        self.backend3.uri_schemes.get.return_value = ['dummy3']
        self.backend3.has_playlists().get.return_value = False
        self.backend3.playlists = None

        self.core = core.Core(mixer=None, backends=[
            self.backend3, self.backend1, self.backend2])


class PlaylistTest(BasePlaylistsTest):

    def test_as_list_combines_result_from_backends(self):
        result = self.core.playlists.as_list()

        self.assertIn(self.plr1a, result)
        self.assertIn(self.plr1b, result)
        self.assertIn(self.plr2a, result)
        self.assertIn(self.plr2b, result)

    def test_as_list_ignores_backends_that_dont_support_it(self):
        self.sp2.as_list.return_value.get.side_effect = NotImplementedError

        result = self.core.playlists.as_list()

        self.assertEqual(len(result), 2)
        self.assertIn(self.plr1a, result)
        self.assertIn(self.plr1b, result)

    def test_get_items_selects_the_matching_backend(self):
        ref = Ref.track()
        self.sp2.get_items.return_value.get.return_value = [ref]

        result = self.core.playlists.get_items('dummy2:pl:a')

        self.assertEqual([ref], result)
        self.assertFalse(self.sp1.get_items.called)
        self.sp2.get_items.assert_called_once_with('dummy2:pl:a')

    def test_get_items_with_unknown_uri_scheme_does_nothing(self):
        result = self.core.playlists.get_items('unknown:a')

        self.assertIsNone(result)
        self.assertFalse(self.sp1.delete.called)
        self.assertFalse(self.sp2.delete.called)

    def test_create_without_uri_scheme_uses_first_backend(self):
        playlist = Playlist()
        self.sp1.create.return_value.get.return_value = playlist

        result = self.core.playlists.create('foo')

        self.assertEqual(playlist, result)
        self.sp1.create.assert_called_once_with('foo')
        self.assertFalse(self.sp2.create.called)

    def test_create_without_uri_scheme_ignores_none_result(self):
        playlist = Playlist()
        self.sp1.create.return_value.get.return_value = None
        self.sp2.create.return_value.get.return_value = playlist

        result = self.core.playlists.create('foo')

        self.assertEqual(playlist, result)
        self.sp1.create.assert_called_once_with('foo')
        self.sp2.create.assert_called_once_with('foo')

    def test_create_without_uri_scheme_ignores_exception(self):
        playlist = Playlist()
        self.sp1.create.return_value.get.side_effect = Exception
        self.sp2.create.return_value.get.return_value = playlist

        result = self.core.playlists.create('foo')

        self.assertEqual(playlist, result)
        self.sp1.create.assert_called_once_with('foo')
        self.sp2.create.assert_called_once_with('foo')

    def test_create_with_uri_scheme_selects_the_matching_backend(self):
        playlist = Playlist()
        self.sp2.create.return_value.get.return_value = playlist

        result = self.core.playlists.create('foo', uri_scheme='dummy2')

        self.assertEqual(playlist, result)
        self.assertFalse(self.sp1.create.called)
        self.sp2.create.assert_called_once_with('foo')

    def test_create_with_unsupported_uri_scheme_uses_first_backend(self):
        playlist = Playlist()
        self.sp1.create.return_value.get.return_value = playlist

        result = self.core.playlists.create('foo', uri_scheme='dummy3')

        self.assertEqual(playlist, result)
        self.sp1.create.assert_called_once_with('foo')
        self.assertFalse(self.sp2.create.called)

    def test_delete_selects_the_dummy1_backend(self):
        success = self.core.playlists.delete('dummy1:a')

        self.assertTrue(success)
        self.sp1.delete.assert_called_once_with('dummy1:a')
        self.assertFalse(self.sp2.delete.called)

    def test_delete_selects_the_dummy2_backend(self):
        success = self.core.playlists.delete('dummy2:a')

        self.assertTrue(success)
        self.assertFalse(self.sp1.delete.called)
        self.sp2.delete.assert_called_once_with('dummy2:a')

    def test_delete_with_unknown_uri_scheme_does_nothing(self):
        success = self.core.playlists.delete('unknown:a')

        self.assertFalse(success)
        self.assertFalse(self.sp1.delete.called)
        self.assertFalse(self.sp2.delete.called)

    def test_delete_ignores_backend_without_playlist_support(self):
        success = self.core.playlists.delete('dummy3:a')

        self.assertFalse(success)
        self.assertFalse(self.sp1.delete.called)
        self.assertFalse(self.sp2.delete.called)

    def test_lookup_selects_the_dummy1_backend(self):
        self.core.playlists.lookup('dummy1:a')

        self.sp1.lookup.assert_called_once_with('dummy1:a')
        self.assertFalse(self.sp2.lookup.called)

    def test_lookup_selects_the_dummy2_backend(self):
        self.core.playlists.lookup('dummy2:a')

        self.assertFalse(self.sp1.lookup.called)
        self.sp2.lookup.assert_called_once_with('dummy2:a')

    def test_lookup_track_in_backend_without_playlists_fails(self):
        result = self.core.playlists.lookup('dummy3:a')

        self.assertIsNone(result)
        self.assertFalse(self.sp1.lookup.called)
        self.assertFalse(self.sp2.lookup.called)

    def test_refresh_without_uri_scheme_refreshes_all_backends(self):
        self.core.playlists.refresh()

        self.sp1.refresh.assert_called_once_with()
        self.sp2.refresh.assert_called_once_with()

    def test_refresh_with_uri_scheme_refreshes_matching_backend(self):
        self.core.playlists.refresh(uri_scheme='dummy2')

        self.assertFalse(self.sp1.refresh.called)
        self.sp2.refresh.assert_called_once_with()

    def test_refresh_with_unknown_uri_scheme_refreshes_nothing(self):
        self.core.playlists.refresh(uri_scheme='foobar')

        self.assertFalse(self.sp1.refresh.called)
        self.assertFalse(self.sp2.refresh.called)

    def test_refresh_ignores_backend_without_playlist_support(self):
        self.core.playlists.refresh(uri_scheme='dummy3')

        self.assertFalse(self.sp1.refresh.called)
        self.assertFalse(self.sp2.refresh.called)

    def test_save_selects_the_dummy1_backend(self):
        playlist = Playlist(uri='dummy1:a')
        self.sp1.save.return_value.get.return_value = playlist

        result = self.core.playlists.save(playlist)

        self.assertEqual(playlist, result)
        self.sp1.save.assert_called_once_with(playlist)
        self.assertFalse(self.sp2.save.called)

    def test_save_selects_the_dummy2_backend(self):
        playlist = Playlist(uri='dummy2:a')
        self.sp2.save.return_value.get.return_value = playlist

        result = self.core.playlists.save(playlist)

        self.assertEqual(playlist, result)
        self.assertFalse(self.sp1.save.called)
        self.sp2.save.assert_called_once_with(playlist)

    def test_save_does_nothing_if_playlist_uri_is_unset(self):
        result = self.core.playlists.save(Playlist())

        self.assertIsNone(result)
        self.assertFalse(self.sp1.save.called)
        self.assertFalse(self.sp2.save.called)

    def test_save_does_nothing_if_playlist_uri_has_unknown_scheme(self):
        result = self.core.playlists.save(Playlist(uri='foobar:a'))

        self.assertIsNone(result)
        self.assertFalse(self.sp1.save.called)
        self.assertFalse(self.sp2.save.called)

    def test_save_ignores_backend_without_playlist_support(self):
        result = self.core.playlists.save(Playlist(uri='dummy3:a'))

        self.assertIsNone(result)
        self.assertFalse(self.sp1.save.called)
        self.assertFalse(self.sp2.save.called)

    def test_get_uri_schemes(self):
        result = self.core.playlists.get_uri_schemes()
        self.assertEquals(result, ['dummy1', 'dummy2'])


class MockBackendCorePlaylistsBase(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.playlists = mock.Mock(spec=backend.PlaylistsProvider)

        self.backend = mock.Mock()
        self.backend.actor_ref.actor_class.__name__ = 'DummyBackend'
        self.backend.uri_schemes.get.return_value = ['dummy']
        self.backend.playlists = self.playlists

        self.core = core.Core(mixer=None, backends=[self.backend])


@mock.patch('mopidy.core.playlists.logger')
class AsListBadBackendsTest(MockBackendCorePlaylistsBase):

    def test_backend_raises_exception(self, logger):
        self.playlists.as_list.return_value.get.side_effect = Exception
        self.assertEqual([], self.core.playlists.as_list())
        logger.exception.assert_called_with(mock.ANY, 'DummyBackend')

    def test_backend_returns_none(self, logger):
        self.playlists.as_list.return_value.get.return_value = None
        self.assertEqual([], self.core.playlists.as_list())
        self.assertFalse(logger.error.called)

    def test_backend_returns_wrong_type(self, logger):
        self.playlists.as_list.return_value.get.return_value = 'abc'
        self.assertEqual([], self.core.playlists.as_list())
        logger.error.assert_called_with(mock.ANY, 'DummyBackend', mock.ANY)


@mock.patch('mopidy.core.playlists.logger')
class GetItemsBadBackendsTest(MockBackendCorePlaylistsBase):

    def test_backend_raises_exception(self, logger):
        self.playlists.get_items.return_value.get.side_effect = Exception
        self.assertIsNone(self.core.playlists.get_items('dummy:/1'))
        logger.exception.assert_called_with(mock.ANY, 'DummyBackend')

    def test_backend_returns_none(self, logger):
        self.playlists.get_items.return_value.get.return_value = None
        self.assertIsNone(self.core.playlists.get_items('dummy:/1'))
        self.assertFalse(logger.error.called)

    def test_backend_returns_wrong_type(self, logger):
        self.playlists.get_items.return_value.get.return_value = 'abc'
        self.assertIsNone(self.core.playlists.get_items('dummy:/1'))
        logger.error.assert_called_with(mock.ANY, 'DummyBackend', mock.ANY)


@mock.patch('mopidy.core.playlists.logger')
class CreateBadBackendsTest(MockBackendCorePlaylistsBase):

    def test_backend_raises_exception(self, logger):
        self.playlists.create.return_value.get.side_effect = Exception
        self.assertIsNone(self.core.playlists.create('foobar'))
        logger.exception.assert_called_with(mock.ANY, 'DummyBackend')

    def test_backend_returns_none(self, logger):
        self.playlists.create.return_value.get.return_value = None
        self.assertIsNone(self.core.playlists.create('foobar'))
        self.assertFalse(logger.error.called)

    def test_backend_returns_wrong_type(self, logger):
        self.playlists.create.return_value.get.return_value = 'abc'
        self.assertIsNone(self.core.playlists.create('foobar'))
        logger.error.assert_called_with(mock.ANY, 'DummyBackend', mock.ANY)


@mock.patch('mopidy.core.playlists.logger')
class DeleteBadBackendsTest(MockBackendCorePlaylistsBase):

    def test_backend_raises_exception(self, logger):
        self.playlists.delete.return_value.get.side_effect = Exception
        self.assertFalse(self.core.playlists.delete('dummy:/1'))
        logger.exception.assert_called_with(mock.ANY, 'DummyBackend')


@mock.patch('mopidy.core.playlists.logger')
class LookupBadBackendsTest(MockBackendCorePlaylistsBase):

    def test_backend_raises_exception(self, logger):
        self.playlists.lookup.return_value.get.side_effect = Exception
        self.assertIsNone(self.core.playlists.lookup('dummy:/1'))
        logger.exception.assert_called_with(mock.ANY, 'DummyBackend')

    def test_backend_returns_none(self, logger):
        self.playlists.lookup.return_value.get.return_value = None
        self.assertIsNone(self.core.playlists.lookup('dummy:/1'))
        self.assertFalse(logger.error.called)

    def test_backend_returns_wrong_type(self, logger):
        self.playlists.lookup.return_value.get.return_value = 'abc'
        self.assertIsNone(self.core.playlists.lookup('dummy:/1'))
        logger.error.assert_called_with(mock.ANY, 'DummyBackend', mock.ANY)


@mock.patch('mopidy.core.playlists.logger')
class RefreshBadBackendsTest(MockBackendCorePlaylistsBase):

    @mock.patch('mopidy.core.listener.CoreListener.send')
    def test_backend_raises_exception(self, send, logger):
        self.playlists.refresh.return_value.get.side_effect = Exception
        self.core.playlists.refresh()
        self.assertFalse(send.called)
        logger.exception.assert_called_with(mock.ANY, 'DummyBackend')

    @mock.patch('mopidy.core.listener.CoreListener.send')
    def test_backend_raises_exception_called_with_uri(self, send, logger):
        self.playlists.refresh.return_value.get.side_effect = Exception
        self.core.playlists.refresh('dummy')
        self.assertFalse(send.called)
        logger.exception.assert_called_with(mock.ANY, 'DummyBackend')


@mock.patch('mopidy.core.playlists.logger')
class SaveBadBackendsTest(MockBackendCorePlaylistsBase):

    def test_backend_raises_exception(self, logger):
        playlist = Playlist(uri='dummy:/1')
        self.playlists.save.return_value.get.side_effect = Exception
        self.assertIsNone(self.core.playlists.save(playlist))
        logger.exception.assert_called_with(mock.ANY, 'DummyBackend')

    def test_backend_returns_none(self, logger):
        playlist = Playlist(uri='dummy:/1')
        self.playlists.save.return_value.get.return_value = None
        self.assertIsNone(self.core.playlists.save(playlist))
        self.assertFalse(logger.error.called)

    def test_backend_returns_wrong_type(self, logger):
        playlist = Playlist(uri='dummy:/1')
        self.playlists.save.return_value.get.return_value = 'abc'
        self.assertIsNone(self.core.playlists.save(playlist))
        logger.error.assert_called_with(mock.ANY, 'DummyBackend', mock.ANY)



from __future__ import absolute_import, unicode_literals

import pykka

from mopidy import backend

from . import playlists


class M3UBackend(pykka.ThreadingActor, backend.Backend):
    uri_schemes = ['m3u']

    def __init__(self, config, audio):
        super(M3UBackend, self).__init__()
        self.playlists = playlists.M3UPlaylistsProvider(self, config)



from __future__ import absolute_import, unicode_literals

import os
import unittest

from mopidy import exceptions
from mopidy.audio import scan
from mopidy.internal import path as path_lib

from tests import path_to_data_dir


class ScannerTest(unittest.TestCase):

    def setUp(self):  # noqa: N802
        self.errors = {}
        self.result = {}

    def find(self, path):
        media_dir = path_to_data_dir(path)
        result, errors = path_lib.find_mtimes(media_dir)
        for path in result:
            yield os.path.join(media_dir, path)

    def scan(self, paths):
        scanner = scan.Scanner()
        for path in paths:
            uri = path_lib.path_to_uri(path)
            key = uri[len('file://'):]
            try:
                self.result[key] = scanner.scan(uri)
            except exceptions.ScannerError as error:
                self.errors[key] = error

    def check(self, name, key, value):
        name = path_to_data_dir(name)
        self.assertEqual(self.result[name].tags[key], value)

    def check_if_missing_plugin(self):
        for path, result in self.result.items():
            if not path.endswith('.mp3'):
                continue
            if not result.playable and result.mime == 'audio/mpeg':
                raise unittest.SkipTest('Missing MP3 support?')

    def test_tags_is_set(self):
        self.scan(self.find('scanner/simple'))

        self.assert_(self.result.values()[0].tags)

    def test_errors_is_not_set(self):
        self.scan(self.find('scanner/simple'))

        self.check_if_missing_plugin()

        self.assert_(not self.errors)

    def test_duration_is_set(self):
        self.scan(self.find('scanner/simple'))

        self.check_if_missing_plugin()

        ogg = path_to_data_dir('scanner/simple/song1.ogg')
        mp3 = path_to_data_dir('scanner/simple/song1.mp3')
        self.assertEqual(self.result[mp3].duration, 4680)
        self.assertEqual(self.result[ogg].duration, 4680)

    def test_artist_is_set(self):
        self.scan(self.find('scanner/simple'))

        self.check_if_missing_plugin()

        self.check('scanner/simple/song1.mp3', 'artist', ['name'])
        self.check('scanner/simple/song1.ogg', 'artist', ['name'])

    def test_album_is_set(self):
        self.scan(self.find('scanner/simple'))

        self.check_if_missing_plugin()

        self.check('scanner/simple/song1.mp3', 'album', ['albumname'])
        self.check('scanner/simple/song1.ogg', 'album', ['albumname'])

    def test_track_is_set(self):
        self.scan(self.find('scanner/simple'))

        self.check_if_missing_plugin()

        self.check('scanner/simple/song1.mp3', 'title', ['trackname'])
        self.check('scanner/simple/song1.ogg', 'title', ['trackname'])

    def test_nonexistant_dir_does_not_fail(self):
        self.scan(self.find('scanner/does-not-exist'))
        self.assert_(not self.errors)

    def test_other_media_is_ignored(self):
        self.scan(self.find('scanner/image'))
        self.assertFalse(self.result.values()[0].playable)

    def test_log_file_that_gst_thinks_is_mpeg_1_is_ignored(self):
        self.scan([path_to_data_dir('scanner/example.log')])

        self.check_if_missing_plugin()

        log = path_to_data_dir('scanner/example.log')
        self.assertLess(self.result[log].duration, 100)

    def test_empty_wav_file(self):
        self.scan([path_to_data_dir('scanner/empty.wav')])
        wav = path_to_data_dir('scanner/empty.wav')
        self.assertEqual(self.result[wav].duration, 0)

    def test_uri_list(self):
        path = path_to_data_dir('scanner/playlist.m3u')
        self.scan([path])
        self.assertEqual(self.result[path].mime, 'text/uri-list')

    def test_text_plain(self):
        # GStreamer decode bin hardcodes bad handling of text plain :/
        path = path_to_data_dir('scanner/plain.txt')
        self.scan([path])
        self.assertIn(path, self.errors)

    @unittest.SkipTest
    def test_song_without_time_is_handeled(self):
        pass



from __future__ import absolute_import, unicode_literals

import copy
import itertools
import weakref

from mopidy.models.fields import Field


# Registered models for automatic deserialization
_models = {}


class ImmutableObject(object):
    """
    Superclass for immutable objects whose fields can only be modified via the
    constructor.

    This version of this class has been retained to avoid breaking any clients
    relying on it's behavior. Internally in Mopidy we now use
    :class:`ValidatedImmutableObject` for type safety and it's much smaller
    memory footprint.

    :param kwargs: kwargs to set as fields on the object
    :type kwargs: any
    """

    # Any sub-classes that don't set slots won't be effected by the base using
    # slots as they will still get an instance dict.
    __slots__ = ['__weakref__']

    def __init__(self, *args, **kwargs):
        for key, value in kwargs.items():
            if not self._is_valid_field(key):
                raise TypeError(
                    '__init__() got an unexpected keyword argument "%s"' % key)
            self._set_field(key, value)

    def __setattr__(self, name, value):
        if name.startswith('_'):
            object.__setattr__(self, name, value)
        else:
            raise AttributeError('Object is immutable.')

    def __delattr__(self, name):
        if name.startswith('_'):
            object.__delattr__(self, name)
        else:
            raise AttributeError('Object is immutable.')

    def _is_valid_field(self, name):
        return hasattr(self, name) and not callable(getattr(self, name))

    def _set_field(self, name, value):
        if value == getattr(self.__class__, name):
            self.__dict__.pop(name, None)
        else:
            self.__dict__[name] = value

    def _items(self):
        return self.__dict__.iteritems()

    def __repr__(self):
        kwarg_pairs = []
        for key, value in sorted(self._items()):
            if isinstance(value, (frozenset, tuple)):
                if not value:
                    continue
                value = list(value)
            kwarg_pairs.append('{}={}'.format(key, repr(value)))
        return '{classname}({kwargs})'.format(
            classname=self.__class__.__name__,
            kwargs=', '.join(kwarg_pairs),
        )

    def __hash__(self):
        hash_sum = 0
        for key, value in self._items():
            hash_sum += hash(key) + hash(value)
        return hash_sum

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return all(a == b for a, b in itertools.izip_longest(
            self._items(), other._items(), fillvalue=object()))

    def __ne__(self, other):
        return not self.__eq__(other)

    def replace(self, **kwargs):
        """
        Replace the fields in the model and return a new instance

        Examples::

            # Returns a track with a new name
            Track(name='foo').replace(name='bar')
            # Return an album with a new number of tracks
            Album(num_tracks=2).replace(num_tracks=5)

        :param kwargs: kwargs to set as fields on the object
        :type kwargs: any
        :rtype: instance of the model with replaced fields
        """
        other = copy.copy(self)
        for key, value in kwargs.items():
            if not self._is_valid_field(key):
                raise TypeError(
                    'replace() got an unexpected keyword argument "%s"' % key)
            other._set_field(key, value)
        return other

    def serialize(self):
        data = {}
        data['__model__'] = self.__class__.__name__
        for key, value in self._items():
            if isinstance(value, (set, frozenset, list, tuple)):
                value = [
                    v.serialize() if isinstance(v, ImmutableObject) else v
                    for v in value]
            elif isinstance(value, ImmutableObject):
                value = value.serialize()
            if not (isinstance(value, list) and len(value) == 0):
                data[key] = value
        return data


class _ValidatedImmutableObjectMeta(type):

    """Helper that initializes fields, slots and memoizes instance creation."""

    def __new__(cls, name, bases, attrs):
        fields = {}

        for base in bases:  # Copy parent fields over to our state
            fields.update(getattr(base, '_fields', {}))

        for key, value in attrs.items():  # Add our own fields
            if isinstance(value, Field):
                fields[key] = '_' + key
                value._name = key

        attrs['_fields'] = fields
        attrs['_instances'] = weakref.WeakValueDictionary()
        attrs['__slots__'] = list(attrs.get('__slots__', [])) + fields.values()

        clsc = super(_ValidatedImmutableObjectMeta, cls).__new__(
            cls, name, bases, attrs)

        if clsc.__name__ != 'ValidatedImmutableObject':
            _models[clsc.__name__] = clsc

        return clsc

    def __call__(cls, *args, **kwargs):  # noqa: N805
        instance = super(_ValidatedImmutableObjectMeta, cls).__call__(
            *args, **kwargs)
        return cls._instances.setdefault(weakref.ref(instance), instance)


class ValidatedImmutableObject(ImmutableObject):
    """
    Superclass for immutable objects whose fields can only be modified via the
    constructor. Fields should be :class:`Field` instances to ensure type
    safety in our models.

    Note that since these models can not be changed, we heavily memoize them
    to save memory. So constructing a class with the same arguments twice will
    give you the same instance twice.
    """

    __metaclass__ = _ValidatedImmutableObjectMeta
    __slots__ = ['_hash']

    def __hash__(self):
        if not hasattr(self, '_hash'):
            hash_sum = super(ValidatedImmutableObject, self).__hash__()
            object.__setattr__(self, '_hash', hash_sum)
        return self._hash

    def _is_valid_field(self, name):
        return name in self._fields

    def _set_field(self, name, value):
        object.__setattr__(self, name, value)

    def _items(self):
        for field, key in self._fields.items():
            if hasattr(self, key):
                yield field, getattr(self, key)

    def replace(self, **kwargs):
        """
        Replace the fields in the model and return a new instance

        Examples::

            # Returns a track with a new name
            Track(name='foo').replace(name='bar')
            # Return an album with a new number of tracks
            Album(num_tracks=2).replace(num_tracks=5)

        Note that internally we memoize heavily to keep memory usage down given
        our overly repetitive data structures. So you might get an existing
        instance if it contains the same values.

        :param kwargs: kwargs to set as fields on the object
        :type kwargs: any
        :rtype: instance of the model with replaced fields
        """
        if not kwargs:
            return self
        other = super(ValidatedImmutableObject, self).replace(**kwargs)
        if hasattr(self, '_hash'):
            object.__delattr__(other, '_hash')
        return self._instances.setdefault(weakref.ref(other), other)



from __future__ import absolute_import, unicode_literals

from mopidy.internal import validation
from mopidy.models import Ref, TlTrack, fields
from mopidy.models.immutable import ValidatedImmutableObject


class HistoryTrack(ValidatedImmutableObject):
    """
    A history track. Wraps a :class:`Ref` and its timestamp.

    :param timestamp: the timestamp
    :type timestamp: int
    :param track: the track reference
    :type track: :class:`Ref`
    """

    # The timestamp. Read-only.
    timestamp = fields.Integer()

    # The track reference. Read-only.
    track = fields.Field(type=Ref)


class HistoryState(ValidatedImmutableObject):
    """
    State of the history controller.
    Internally used for save/load state.

    :param history: the track history
    :type history: list of :class:`HistoryTrack`
    """

    # The tracks. Read-only.
    history = fields.Collection(type=HistoryTrack, container=tuple)


class MixerState(ValidatedImmutableObject):
    """
    State of the mixer controller.
    Internally used for save/load state.

    :param volume: the volume
    :type volume: int
    :param mute: the mute state
    :type mute: int
    """

    # The volume. Read-only.
    volume = fields.Integer(min=0, max=100)

    # The mute state. Read-only.
    mute = fields.Boolean(default=False)


class PlaybackState(ValidatedImmutableObject):
    """
    State of the playback controller.
    Internally used for save/load state.

    :param tlid: current track tlid
    :type tlid: int
    :param time_position: play position
    :type time_position: int
    :param state: playback state
    :type state: :class:`validation.PLAYBACK_STATES`
    """

    # The tlid of current playing track. Read-only.
    tlid = fields.Integer(min=1)

    # The playback position. Read-only.
    time_position = fields.Integer(min=0)

    # The playback state. Read-only.
    state = fields.Field(choices=validation.PLAYBACK_STATES)


class TracklistState(ValidatedImmutableObject):

    """
    State of the tracklist controller.
    Internally used for save/load state.

    :param repeat: the repeat mode
    :type repeat: bool
    :param consume: the consume mode
    :type consume: bool
    :param random: the random mode
    :type random: bool
    :param single: the single mode
    :type single: bool
    :param next_tlid: the id for the next added track
    :type next_tlid: int
    :param tl_tracks: the list of tracks
    :type tl_tracks: list of :class:`TlTrack`
    """

    # The repeat mode. Read-only.
    repeat = fields.Boolean()

    # The consume mode. Read-only.
    consume = fields.Boolean()

    # The random mode. Read-only.
    random = fields.Boolean()

    # The single mode. Read-only.
    single = fields.Boolean()

    # The id of the track to play. Read-only.
    next_tlid = fields.Integer(min=0)

    # The list of tracks. Read-only.
    tl_tracks = fields.Collection(type=TlTrack, container=tuple)


class CoreState(ValidatedImmutableObject):

    """
    State of all Core controller.
    Internally used for save/load state.

    :param history: State of the history controller
    :type history: :class:`HistorState`
    :param mixer: State of the mixer controller
    :type mixer: :class:`MixerState`
    :param playback: State of the playback controller
    :type playback: :class:`PlaybackState`
    :param tracklist: State of the tracklist controller
    :type tracklist: :class:`TracklistState`
    """

    # State of the history controller.
    history = fields.Field(type=HistoryState)

    # State of the mixer controller.
    mixer = fields.Field(type=MixerState)

    # State of the playback controller.
    playback = fields.Field(type=PlaybackState)

    # State of the tracklist controller.
    tracklist = fields.Field(type=TracklistState)



from __future__ import absolute_import, unicode_literals

import json
import unittest

from mopidy.internal.models import (
    HistoryState, HistoryTrack, MixerState, PlaybackState, TracklistState)
from mopidy.models import (
    ModelJSONEncoder, Ref, TlTrack, Track, model_json_decoder)


class HistoryTrackTest(unittest.TestCase):

    def test_track(self):
        track = Ref.track()
        result = HistoryTrack(track=track)
        self.assertEqual(result.track, track)
        with self.assertRaises(AttributeError):
            result.track = None

    def test_timestamp(self):
        timestamp = 1234
        result = HistoryTrack(timestamp=timestamp)
        self.assertEqual(result.timestamp, timestamp)
        with self.assertRaises(AttributeError):
            result.timestamp = None

    def test_to_json_and_back(self):
        result = HistoryTrack(track=Ref.track(), timestamp=1234)
        serialized = json.dumps(result, cls=ModelJSONEncoder)
        deserialized = json.loads(serialized, object_hook=model_json_decoder)
        self.assertEqual(result, deserialized)


class HistoryStateTest(unittest.TestCase):

    def test_history_list(self):
        history = (HistoryTrack(),
                   HistoryTrack())
        result = HistoryState(history=history)
        self.assertEqual(result.history, history)
        with self.assertRaises(AttributeError):
            result.history = None

    def test_history_string_fail(self):
        history = 'not_a_valid_history'
        with self.assertRaises(TypeError):
            HistoryState(history=history)

    def test_to_json_and_back(self):
        result = HistoryState(history=(HistoryTrack(), HistoryTrack()))
        serialized = json.dumps(result, cls=ModelJSONEncoder)
        deserialized = json.loads(serialized, object_hook=model_json_decoder)
        self.assertEqual(result, deserialized)


class MixerStateTest(unittest.TestCase):

    def test_volume(self):
        volume = 37
        result = MixerState(volume=volume)
        self.assertEqual(result.volume, volume)
        with self.assertRaises(AttributeError):
            result.volume = None

    def test_volume_invalid(self):
        volume = 105
        with self.assertRaises(ValueError):
            MixerState(volume=volume)

    def test_mute_false(self):
        mute = False
        result = MixerState(mute=mute)
        self.assertEqual(result.mute, mute)
        with self.assertRaises(AttributeError):
            result.mute = None

    def test_mute_true(self):
        mute = True
        result = MixerState(mute=mute)
        self.assertEqual(result.mute, mute)
        with self.assertRaises(AttributeError):
            result.mute = False

    def test_mute_default(self):
        result = MixerState()
        self.assertEqual(result.mute, False)

    def test_to_json_and_back(self):
        result = MixerState(volume=77)
        serialized = json.dumps(result, cls=ModelJSONEncoder)
        deserialized = json.loads(serialized, object_hook=model_json_decoder)
        self.assertEqual(result, deserialized)


class PlaybackStateTest(unittest.TestCase):

    def test_position(self):
        time_position = 123456
        result = PlaybackState(time_position=time_position)
        self.assertEqual(result.time_position, time_position)
        with self.assertRaises(AttributeError):
            result.time_position = None

    def test_position_invalid(self):
        time_position = -1
        with self.assertRaises(ValueError):
            PlaybackState(time_position=time_position)

    def test_tl_track(self):
        tlid = 42
        result = PlaybackState(tlid=tlid)
        self.assertEqual(result.tlid, tlid)
        with self.assertRaises(AttributeError):
            result.tlid = None

    def test_tl_track_none(self):
        tlid = None
        result = PlaybackState(tlid=tlid)
        self.assertEqual(result.tlid, tlid)
        with self.assertRaises(AttributeError):
            result.tl_track = None

    def test_tl_track_invalid(self):
        tl_track = Track()
        with self.assertRaises(TypeError):
            PlaybackState(tlid=tl_track)

    def test_state(self):
        state = 'playing'
        result = PlaybackState(state=state)
        self.assertEqual(result.state, state)
        with self.assertRaises(AttributeError):
            result.state = None

    def test_state_invalid(self):
        state = 'not_a_state'
        with self.assertRaises(TypeError):
            PlaybackState(state=state)

    def test_to_json_and_back(self):
        result = PlaybackState(state='playing', tlid=4321)
        serialized = json.dumps(result, cls=ModelJSONEncoder)
        deserialized = json.loads(serialized, object_hook=model_json_decoder)
        self.assertEqual(result, deserialized)


class TracklistStateTest(unittest.TestCase):

    def test_repeat_true(self):
        repeat = True
        result = TracklistState(repeat=repeat)
        self.assertEqual(result.repeat, repeat)
        with self.assertRaises(AttributeError):
            result.repeat = None

    def test_repeat_false(self):
        repeat = False
        result = TracklistState(repeat=repeat)
        self.assertEqual(result.repeat, repeat)
        with self.assertRaises(AttributeError):
            result.repeat = None

    def test_repeat_invalid(self):
        repeat = 33
        with self.assertRaises(TypeError):
            TracklistState(repeat=repeat)

    def test_consume_true(self):
        val = True
        result = TracklistState(consume=val)
        self.assertEqual(result.consume, val)
        with self.assertRaises(AttributeError):
            result.repeat = None

    def test_random_true(self):
        val = True
        result = TracklistState(random=val)
        self.assertEqual(result.random, val)
        with self.assertRaises(AttributeError):
            result.random = None

    def test_single_true(self):
        val = True
        result = TracklistState(single=val)
        self.assertEqual(result.single, val)
        with self.assertRaises(AttributeError):
            result.single = None

    def test_next_tlid(self):
        val = 654
        result = TracklistState(next_tlid=val)
        self.assertEqual(result.next_tlid, val)
        with self.assertRaises(AttributeError):
            result.next_tlid = None

    def test_next_tlid_invalid(self):
        val = -1
        with self.assertRaises(ValueError):
            TracklistState(next_tlid=val)

    def test_tracks(self):
        tracks = (TlTrack(), TlTrack())
        result = TracklistState(tl_tracks=tracks)
        self.assertEqual(result.tl_tracks, tracks)
        with self.assertRaises(AttributeError):
            result.tl_tracks = None

    def test_tracks_invalid(self):
        tracks = (Track(), Track())
        with self.assertRaises(TypeError):
            TracklistState(tl_tracks=tracks)

    def test_to_json_and_back(self):
        result = TracklistState(tl_tracks=(TlTrack(), TlTrack()), next_tlid=4)
        serialized = json.dumps(result, cls=ModelJSONEncoder)
        deserialized = json.loads(serialized, object_hook=model_json_decoder)
        self.assertEqual(result, deserialized)



from __future__ import absolute_import, unicode_literals

import locale

from mopidy import compat


def locale_decode(bytestr):
    try:
        return compat.text_type(bytestr)
    except UnicodeError:
        return bytes(bytestr).decode(locale.getpreferredencoding(), 'replace')



from __future__ import absolute_import, unicode_literals

from tests.mpd import protocol


class AudioOutputHandlerTest(protocol.BaseTestCase):

    def test_enableoutput(self):
        self.core.mixer.set_mute(False)

        self.send_request('enableoutput "0"')

        self.assertInResponse('OK')
        self.assertEqual(self.core.mixer.get_mute().get(), True)

    def test_enableoutput_unknown_outputid(self):
        self.send_request('enableoutput "7"')

        self.assertInResponse('ACK [50@0] {enableoutput} No such audio output')

    def test_disableoutput(self):
        self.core.mixer.set_mute(True)

        self.send_request('disableoutput "0"')

        self.assertInResponse('OK')
        self.assertEqual(self.core.mixer.get_mute().get(), False)

    def test_disableoutput_unknown_outputid(self):
        self.send_request('disableoutput "7"')

        self.assertInResponse(
            'ACK [50@0] {disableoutput} No such audio output')

    def test_outputs_when_unmuted(self):
        self.core.mixer.set_mute(False)

        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 0')
        self.assertInResponse('OK')

    def test_outputs_when_muted(self):
        self.core.mixer.set_mute(True)

        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 1')
        self.assertInResponse('OK')

    def test_outputs_toggleoutput(self):
        self.core.mixer.set_mute(False)

        self.send_request('toggleoutput "0"')
        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 1')
        self.assertInResponse('OK')

        self.send_request('toggleoutput "0"')
        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 0')
        self.assertInResponse('OK')

        self.send_request('toggleoutput "0"')
        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 1')
        self.assertInResponse('OK')

    def test_outputs_toggleoutput_unknown_outputid(self):
        self.send_request('toggleoutput "7"')

        self.assertInResponse(
            'ACK [50@0] {toggleoutput} No such audio output')


class AudioOutputHandlerNoneMixerTest(protocol.BaseTestCase):
    enable_mixer = False

    def test_enableoutput(self):
        self.assertEqual(self.core.mixer.get_mute().get(), None)

        self.send_request('enableoutput "0"')
        self.assertInResponse(
            'ACK [52@0] {enableoutput} problems enabling output')

        self.assertEqual(self.core.mixer.get_mute().get(), None)

    def test_disableoutput(self):
        self.assertEqual(self.core.mixer.get_mute().get(), None)

        self.send_request('disableoutput "0"')
        self.assertInResponse(
            'ACK [52@0] {disableoutput} problems disabling output')

        self.assertEqual(self.core.mixer.get_mute().get(), None)

    def test_outputs_when_unmuted(self):
        self.core.mixer.set_mute(False)

        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 0')
        self.assertInResponse('OK')

    def test_outputs_when_muted(self):
        self.core.mixer.set_mute(True)

        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 0')
        self.assertInResponse('OK')

    def test_outputs_toggleoutput(self):
        self.core.mixer.set_mute(False)

        self.send_request('toggleoutput "0"')
        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 0')
        self.assertInResponse('OK')

        self.send_request('toggleoutput "0"')
        self.send_request('outputs')

        self.assertInResponse('outputid: 0')
        self.assertInResponse('outputname: Mute')
        self.assertInResponse('outputenabled: 0')
        self.assertInResponse('OK')



from __future__ import absolute_import, unicode_literals

import mock

import pytest

import requests

import responses

from mopidy.internal import http


TIMEOUT = 1000
URI = 'http://example.com/foo.txt'
BODY = "This is the contents of foo.txt."


@pytest.fixture
def session():
    return requests.Session()


@pytest.fixture
def session_mock():
    return mock.Mock(spec=requests.Session)


@responses.activate
def test_download_on_server_side_error(session, caplog):
    responses.add(responses.GET, URI, body=BODY, status=500)

    result = http.download(session, URI)

    assert result is None
    assert 'Problem downloading' in caplog.text


def test_download_times_out_if_connection_times_out(session_mock, caplog):
    session_mock.get.side_effect = requests.exceptions.Timeout

    result = http.download(session_mock, URI, timeout=1.0)

    session_mock.get.assert_called_once_with(URI, timeout=1.0, stream=True)
    assert result is None
    assert (
        'Download of %r failed due to connection timeout after 1.000s' % URI
        in caplog.text)


@responses.activate
def test_download_times_out_if_download_is_slow(session, caplog):
    responses.add(responses.GET, URI, body=BODY, content_type='text/plain')

    with mock.patch.object(http, 'time') as time_mock:
        time_mock.time.side_effect = [0, TIMEOUT + 1]

        result = http.download(session, URI)

    assert result is None
    assert (
        'Download of %r failed due to download taking more than 1.000s' % URI
        in caplog.text)



from __future__ import absolute_import, unicode_literals

import logging
import os

import mopidy
from mopidy import config, ext

logger = logging.getLogger(__name__)


class Extension(ext.Extension):

    dist_name = 'Mopidy-File'
    ext_name = 'file'
    version = mopidy.__version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), 'ext.conf')
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super(Extension, self).get_config_schema()
        schema['media_dirs'] = config.List(optional=True)
        schema['excluded_file_extensions'] = config.List(optional=True)
        schema['show_dotfiles'] = config.Boolean(optional=True)
        schema['follow_symlinks'] = config.Boolean(optional=True)
        schema['metadata_timeout'] = config.Integer(optional=True)
        return schema

    def setup(self, registry):
        from .backend import FileBackend
        registry.add('backend', FileBackend)



from __future__ import absolute_import, unicode_literals

import logging
import time

import requests

from mopidy import httpclient

logger = logging.getLogger(__name__)


def get_requests_session(proxy_config, user_agent):
    proxy = httpclient.format_proxy(proxy_config)
    full_user_agent = httpclient.format_user_agent(user_agent)

    session = requests.Session()
    session.proxies.update({'http': proxy, 'https': proxy})
    session.headers.update({'user-agent': full_user_agent})

    return session


def download(session, uri, timeout=1.0, chunk_size=4096):
    try:
        response = session.get(uri, stream=True, timeout=timeout)
    except requests.exceptions.Timeout:
        logger.warning('Download of %r failed due to connection timeout after '
                       '%.3fs', uri, timeout)
        return None
    except requests.exceptions.InvalidSchema:
        logger.warning('Download of %r failed due to unsupported schema', uri)
        return None
    except requests.exceptions.RequestException as exc:
        logger.warning('Download of %r failed: %s', uri, exc)
        logger.debug('Download exception details', exc_info=True)
        return None

    content = []
    deadline = time.time() + timeout
    for chunk in response.iter_content(chunk_size):
        content.append(chunk)
        if time.time() > deadline:
            logger.warning('Download of %r failed due to download taking more '
                           'than %.3fs', uri, timeout)
            return None

    if not response.ok:
        logger.warning('Problem downloading %r: %s', uri, response.reason)
        return None

    return b''.join(content)



from __future__ import absolute_import, unicode_literals

import contextlib
import io
import locale
import logging
import operator
import os
import tempfile

from mopidy import backend
from mopidy.internal import path

from . import Extension, translator

logger = logging.getLogger(__name__)


def log_environment_error(message, error):
    if isinstance(error.strerror, bytes):
        strerror = error.strerror.decode(locale.getpreferredencoding())
    else:
        strerror = error.strerror
    logger.error('%s: %s', message, strerror)


@contextlib.contextmanager
def replace(path, mode='w+b', encoding=None, errors=None):
    try:
        (fd, tempname) = tempfile.mkstemp(dir=os.path.dirname(path))
    except TypeError:
        # Python 3 requires dir to be of type str until v3.5
        import sys
        path = path.decode(sys.getfilesystemencoding())
        (fd, tempname) = tempfile.mkstemp(dir=os.path.dirname(path))
    try:
        fp = io.open(fd, mode, encoding=encoding, errors=errors)
    except Exception:
        os.remove(tempname)
        os.close(fd)
        raise
    try:
        yield fp
        fp.flush()
        os.fsync(fd)
        os.rename(tempname, path)
    except Exception:
        os.remove(tempname)
        raise
    finally:
        fp.close()


class M3UPlaylistsProvider(backend.PlaylistsProvider):

    def __init__(self, backend, config):
        super(M3UPlaylistsProvider, self).__init__(backend)

        ext_config = config[Extension.ext_name]
        if ext_config['playlists_dir'] is None:
            self._playlists_dir = Extension.get_data_dir(config)
        else:
            self._playlists_dir = ext_config['playlists_dir']
        self._base_dir = ext_config['base_dir'] or self._playlists_dir
        self._default_encoding = ext_config['default_encoding']
        self._default_extension = ext_config['default_extension']

    def as_list(self):
        result = []
        for entry in os.listdir(self._playlists_dir):
            if not entry.endswith((b'.m3u', b'.m3u8')):
                continue
            elif not os.path.isfile(self._abspath(entry)):
                continue
            else:
                result.append(translator.path_to_ref(entry))
        result.sort(key=operator.attrgetter('name'))
        return result

    def create(self, name):
        path = translator.path_from_name(name.strip(), self._default_extension)
        try:
            with self._open(path, 'w'):
                pass
            mtime = os.path.getmtime(self._abspath(path))
        except EnvironmentError as e:
            log_environment_error('Error creating playlist %s' % name, e)
        else:
            return translator.playlist(path, [], mtime)

    def delete(self, uri):
        path = translator.uri_to_path(uri)
        if not self._is_in_basedir(path):
            logger.debug('Ignoring path outside playlist dir: %s', uri)
            return False
        try:
            os.remove(self._abspath(path))
        except EnvironmentError as e:
            log_environment_error('Error deleting playlist %s' % uri, e)
            return False
        else:
            return True

    def get_items(self, uri):
        path = translator.uri_to_path(uri)
        if not self._is_in_basedir(path):
            logger.debug('Ignoring path outside playlist dir: %s', uri)
            return None
        try:
            with self._open(path, 'r') as fp:
                items = translator.load_items(fp, self._base_dir)
        except EnvironmentError as e:
            log_environment_error('Error reading playlist %s' % uri, e)
        else:
            return items

    def lookup(self, uri):
        path = translator.uri_to_path(uri)
        if not self._is_in_basedir(path):
            logger.debug('Ignoring path outside playlist dir: %s', uri)
            return None
        try:
            with self._open(path, 'r') as fp:
                items = translator.load_items(fp, self._base_dir)
            mtime = os.path.getmtime(self._abspath(path))
        except EnvironmentError as e:
            log_environment_error('Error reading playlist %s' % uri, e)
        else:
            return translator.playlist(path, items, mtime)

    def refresh(self):
        pass  # nothing to do

    def save(self, playlist):
        path = translator.uri_to_path(playlist.uri)
        if not self._is_in_basedir(path):
            logger.debug(
                'Ignoring path outside playlist dir: %s', playlist.uri)
            return None
        name = translator.name_from_path(path)
        try:
            with self._open(path, 'w') as fp:
                translator.dump_items(playlist.tracks, fp)
            if playlist.name and playlist.name != name:
                opath, ext = os.path.splitext(path)
                path = translator.path_from_name(playlist.name.strip()) + ext
                os.rename(self._abspath(opath + ext), self._abspath(path))
            mtime = os.path.getmtime(self._abspath(path))
        except EnvironmentError as e:
            log_environment_error('Error saving playlist %s' % playlist.uri, e)
        else:
            return translator.playlist(path, playlist.tracks, mtime)

    def _abspath(self, path):
        return os.path.join(self._playlists_dir, path)

    def _is_in_basedir(self, local_path):
        if not os.path.isabs(local_path):
            local_path = os.path.join(self._playlists_dir, local_path)
        return path.is_path_inside_base_dir(local_path, self._playlists_dir)

    def _open(self, path, mode='r'):
        if path.endswith(b'.m3u8'):
            encoding = 'utf-8'
        else:
            encoding = self._default_encoding
        if not os.path.isabs(path):
            path = os.path.join(self._playlists_dir, path)
        if not self._is_in_basedir(path):
            raise Exception(
                'Path (%s) is not inside playlist_dir (%s)'
                % (path, self._playlists_dir))
        if 'w' in mode:
            return replace(path, mode, encoding=encoding, errors='replace')
        else:
            return io.open(path, mode, encoding=encoding, errors='replace')



from __future__ import absolute_import, unicode_literals

from mopidy.mpd import exceptions, protocol


@protocol.commands.add('mount')
def mount(context, path, uri):
    """
    *musicpd.org, mounts and neighbors section:*

        ``mount {PATH} {URI}``

        Mount the specified remote storage URI at the given path. Example::

            mount foo nfs://192.168.1.4/export/mp3

    .. versionadded:: 0.19
        New in MPD protocol version 0.19
    """
    raise exceptions.MpdNotImplemented  # TODO


@protocol.commands.add('unmount')
def unmount(context, path):
    """
    *musicpd.org, mounts and neighbors section:*

        ``unmount {PATH}``

        Unmounts the specified path. Example::

            unmount foo

    .. versionadded:: 0.19
        New in MPD protocol version 0.19
    """
    raise exceptions.MpdNotImplemented  # TODO


@protocol.commands.add('listmounts')
def listmounts(context):
    """
    *musicpd.org, mounts and neighbors section:*

        ``listmounts``

        Queries a list of all mounts. By default, this contains just the
        configured music_directory. Example::

            listmounts
            mount:
            storage: /home/foo/music
            mount: foo
            storage: nfs://192.168.1.4/export/mp3
            OK

    .. versionadded:: 0.19
        New in MPD protocol version 0.19
    """
    raise exceptions.MpdNotImplemented  # TODO


@protocol.commands.add('listneighbors')
def listneighbors(context):
    """
    *musicpd.org, mounts and neighbors section:*

        ``listneighbors``

        Queries a list of "neighbors" (e.g. accessible file servers on the
        local net). Items on that list may be used with the mount command.
        Example::

            listneighbors
            neighbor: smb://FOO
            name: FOO (Samba 4.1.11-Debian)
            OK

    .. versionadded:: 0.19
        New in MPD protocol version 0.19
    """
    raise exceptions.MpdNotImplemented  # TODO



from __future__ import absolute_import, unicode_literals

import random

import mock

from mopidy.models import Playlist, Ref, Track
from mopidy.mpd.protocol import stored_playlists

from tests.mpd import protocol


class IssueGH17RegressionTest(protocol.BaseTestCase):

    """
    The issue: http://github.com/mopidy/mopidy/issues/17

    How to reproduce:

    - Play a playlist where one track cannot be played
    - Turn on random mode
    - Press next until you get to the unplayable track
    """

    def test(self):
        tracks = [
            Track(uri='dummy:a'),
            Track(uri='dummy:b'),
            Track(uri='dummy:error'),
            Track(uri='dummy:d'),
            Track(uri='dummy:e'),
            Track(uri='dummy:f'),
        ]
        self.audio.trigger_fake_playback_failure('dummy:error')
        self.backend.library.dummy_library = tracks
        self.core.tracklist.add(uris=[t.uri for t in tracks]).get()

        random.seed(1)  # Playlist order: abcfde

        self.send_request('play')
        self.assertEqual(
            'dummy:a', self.core.playback.get_current_track().get().uri)
        self.send_request('random "1"')
        self.send_request('next')
        self.assertEqual(
            'dummy:b', self.core.playback.get_current_track().get().uri)
        self.send_request('next')
        # Should now be at track 'c', but playback fails and it skips ahead
        self.assertEqual(
            'dummy:f', self.core.playback.get_current_track().get().uri)
        self.send_request('next')
        self.assertEqual(
            'dummy:d', self.core.playback.get_current_track().get().uri)
        self.send_request('next')
        self.assertEqual(
            'dummy:e', self.core.playback.get_current_track().get().uri)


class IssueGH18RegressionTest(protocol.BaseTestCase):

    """
    The issue: http://github.com/mopidy/mopidy/issues/18

    How to reproduce:

        Play, random on, next, random off, next, next.

        At this point it gives the same song over and over.
    """

    def test(self):
        tracks = [
            Track(uri='dummy:a'), Track(uri='dummy:b'), Track(uri='dummy:c'),
            Track(uri='dummy:d'), Track(uri='dummy:e'), Track(uri='dummy:f'),
        ]
        self.backend.library.dummy_library = tracks
        self.core.tracklist.add(uris=[t.uri for t in tracks]).get()

        random.seed(1)

        self.send_request('play')
        self.send_request('random "1"')
        self.send_request('next')
        self.send_request('random "0"')
        self.send_request('next')

        self.send_request('next')
        tl_track_1 = self.core.playback.get_current_tl_track().get()
        self.send_request('next')
        tl_track_2 = self.core.playback.get_current_tl_track().get()
        self.send_request('next')
        tl_track_3 = self.core.playback.get_current_tl_track().get()

        self.assertNotEqual(tl_track_1, tl_track_2)
        self.assertNotEqual(tl_track_2, tl_track_3)


class IssueGH22RegressionTest(protocol.BaseTestCase):

    """
    The issue: http://github.com/mopidy/mopidy/issues/22

    How to reproduce:

        Play, random on, remove all tracks from the current playlist (as in
        "delete" each one, not "clear").

        Alternatively: Play, random on, remove a random track from the current
        playlist, press next until it crashes.
    """

    def test(self):
        tracks = [
            Track(uri='dummy:a'), Track(uri='dummy:b'), Track(uri='dummy:c'),
            Track(uri='dummy:d'), Track(uri='dummy:e'), Track(uri='dummy:f'),
        ]
        self.backend.library.dummy_library = tracks
        self.core.tracklist.add(uris=[t.uri for t in tracks]).get()

        random.seed(1)

        self.send_request('play')
        self.send_request('random "1"')
        self.send_request('deleteid "1"')
        self.send_request('deleteid "2"')
        self.send_request('deleteid "3"')
        self.send_request('deleteid "4"')
        self.send_request('deleteid "5"')
        self.send_request('deleteid "6"')
        self.send_request('status')


class IssueGH69RegressionTest(protocol.BaseTestCase):

    """
    The issue: https://github.com/mopidy/mopidy/issues/69

    How to reproduce:

        Play track, stop, clear current playlist, load a new playlist, status.

        The status response now contains "song: None".
    """

    def test(self):
        self.core.playlists.create('foo')

        tracks = [
            Track(uri='dummy:a'), Track(uri='dummy:b'), Track(uri='dummy:c'),
            Track(uri='dummy:d'), Track(uri='dummy:e'), Track(uri='dummy:f'),
        ]
        self.backend.library.dummy_library = tracks
        self.core.tracklist.add(uris=[t.uri for t in tracks]).get()

        self.send_request('play')
        self.send_request('stop')
        self.send_request('clear')
        self.send_request('load "foo"')
        self.assertNotInResponse('song: None')


class IssueGH113RegressionTest(protocol.BaseTestCase):

    r"""
    The issue: https://github.com/mopidy/mopidy/issues/113

    How to reproduce:

    - Have a playlist with a name contining backslashes, like
      "all lart spotify:track:\w\{22\} pastes".
    - Try to load the playlist with the backslashes in the playlist name
      escaped.
    """

    def test(self):
        self.core.playlists.create(
            r'all lart spotify:track:\w\{22\} pastes')

        self.send_request('lsinfo "/"')
        self.assertInResponse(
            r'playlist: all lart spotify:track:\w\{22\} pastes')

        self.send_request(
            r'listplaylistinfo "all lart spotify:track:\\w\\{22\\} pastes"')
        self.assertInResponse('OK')


class IssueGH137RegressionTest(protocol.BaseTestCase):

    """
    The issue: https://github.com/mopidy/mopidy/issues/137

    How to reproduce:

    - Send "list" query with mismatching quotes
    """

    def test(self):
        self.send_request(
            'list Date Artist "Anita Ward" '
            'Album "This Is Remixed Hits - Mashups & Rare 12" Mixes"')

        self.assertInResponse('ACK [2@0] {list} Invalid unquoted character')


class IssueGH1120RegressionTest(protocol.BaseTestCase):
    """
    The issue: https://github.com/mopidy/mopidy/issues/1120

    How to reproduce:

    - A playlist must be in both browse results and playlists
    - Call for instance ``lsinfo "/"`` to populate the cache with the
      playlist name from the playlist backend.
    - Call ``lsinfo "/dummy"`` to override the playlist name with the browse
      name.
    - Call ``lsinfo "/"`` and we now have an invalid name with ``/`` in it.

    """

    @mock.patch.object(stored_playlists, '_get_last_modified')
    def test(self, last_modified_mock):
        last_modified_mock.return_value = '2015-08-05T22:51:06Z'
        self.backend.library.dummy_browse_result = {
            'dummy:/': [Ref.playlist(name='Top 100 tracks', uri='dummy:/1')],
        }
        self.backend.playlists.set_dummy_playlists([
            Playlist(name='Top 100 tracks', uri='dummy:/1'),
        ])

        response1 = self.send_request('lsinfo "/"')
        self.send_request('lsinfo "/dummy"')

        response2 = self.send_request('lsinfo "/"')
        self.assertEqual(response1, response2)


class IssueGH1348RegressionTest(protocol.BaseTestCase):

    """
    The issue: http://github.com/mopidy/mopidy/issues/1348
    """

    def test(self):
        self.backend.library.dummy_library = [Track(uri='dummy:a')]

        # Create a dummy playlist and trigger population of mapping
        self.send_request('playlistadd "testing1" "dummy:a"')
        self.send_request('listplaylists')

        # Create an other playlist which isn't in the map
        self.send_request('playlistadd "testing2" "dummy:a"')
        self.assertEqual(['OK'], self.send_request('rm "testing2"'))

        playlists = self.backend.playlists.as_list().get()
        self.assertEqual(['testing1'], [ref.name for ref in playlists])



from __future__ import absolute_import, unicode_literals

from mopidy import config


def test_core_schema_has_cache_dir():
    assert 'cache_dir' in config._core_schema
    assert isinstance(config._core_schema['cache_dir'], config.Path)


def test_core_schema_has_config_dir():
    assert 'config_dir' in config._core_schema
    assert isinstance(config._core_schema['config_dir'], config.Path)


def test_core_schema_has_data_dir():
    assert 'data_dir' in config._core_schema
    assert isinstance(config._core_schema['data_dir'], config.Path)


def test_core_schema_has_max_tracklist_length():
    assert 'max_tracklist_length' in config._core_schema
    max_tracklist_length_schema = config._core_schema['max_tracklist_length']
    assert isinstance(max_tracklist_length_schema, config.Integer)
    assert max_tracklist_length_schema._minimum == 1



"""A dummy audio actor for use in tests.

This class implements the audio API in the simplest way possible. It is used in
tests of the core and backends.
"""

from __future__ import absolute_import, unicode_literals

import pykka

from mopidy import audio


def create_proxy(config=None, mixer=None):
    return DummyAudio.start(config, mixer).proxy()


# TODO: reset position on track change?
class DummyAudio(pykka.ThreadingActor):

    def __init__(self, config=None, mixer=None):
        super(DummyAudio, self).__init__()
        self.state = audio.PlaybackState.STOPPED
        self._volume = 0
        self._position = 0
        self._callback = None
        self._uri = None
        self._stream_changed = False
        self._tags = {}
        self._bad_uris = set()

    def set_uri(self, uri):
        assert self._uri is None, 'prepare change not called before set'
        self._position = 0
        self._uri = uri
        self._stream_changed = True
        self._tags = {}

    def set_appsrc(self, *args, **kwargs):
        pass

    def emit_data(self, buffer_):
        pass

    def get_position(self):
        return self._position

    def set_position(self, position):
        self._position = position
        audio.AudioListener.send('position_changed', position=position)
        return True

    def start_playback(self):
        return self._change_state(audio.PlaybackState.PLAYING)

    def pause_playback(self):
        return self._change_state(audio.PlaybackState.PAUSED)

    def prepare_change(self):
        self._uri = None
        return True

    def stop_playback(self):
        return self._change_state(audio.PlaybackState.STOPPED)

    def get_volume(self):
        return self._volume

    def set_volume(self, volume):
        self._volume = volume
        return True

    def set_metadata(self, track):
        pass

    def get_current_tags(self):
        return self._tags

    def set_about_to_finish_callback(self, callback):
        self._callback = callback

    def enable_sync_handler(self):
        pass

    def wait_for_state_change(self):
        pass

    def _change_state(self, new_state):
        if not self._uri:
            return False

        if new_state == audio.PlaybackState.STOPPED and self._uri:
            self._stream_changed = True
            self._uri = None

        if self._uri is not None:
            audio.AudioListener.send('position_changed', position=0)

        if self._stream_changed:
            self._stream_changed = False
            audio.AudioListener.send('stream_changed', uri=self._uri)

        old_state, self.state = self.state, new_state
        audio.AudioListener.send(
            'state_changed',
            old_state=old_state, new_state=new_state, target_state=None)

        if new_state == audio.PlaybackState.PLAYING:
            self._tags['audio-codec'] = ['fake info...']
            audio.AudioListener.send('tags_changed', tags=['audio-codec'])

        return self._uri not in self._bad_uris

    def trigger_fake_playback_failure(self, uri):
        self._bad_uris.add(uri)

    def trigger_fake_tags_changed(self, tags):
        self._tags.update(tags)
        audio.AudioListener.send('tags_changed', tags=self._tags.keys())

    def get_about_to_finish_callback(self):
        # This needs to be called from outside the actor or we lock up.
        def wrapper():
            if self._callback:
                self.prepare_change()
                self._callback()

            if not self._uri or not self._callback:
                self._tags = {}
                audio.AudioListener.send('reached_end_of_stream')
            else:
                audio.AudioListener.send('position_changed', position=0)
                audio.AudioListener.send('stream_changed', uri=self._uri)

        return wrapper



from __future__ import absolute_import, unicode_literals

import contextlib
import logging

from mopidy import exceptions
from mopidy.internal import validation
from mopidy.internal.models import MixerState


logger = logging.getLogger(__name__)


@contextlib.contextmanager
def _mixer_error_handling(mixer):
    try:
        yield
    except exceptions.ValidationError as e:
        logger.error('%s mixer returned bad data: %s',
                     mixer.actor_ref.actor_class.__name__, e)
    except Exception:
        logger.exception('%s mixer caused an exception.',
                         mixer.actor_ref.actor_class.__name__)


class MixerController(object):
    pykka_traversable = True

    def __init__(self, mixer):
        self._mixer = mixer

    def get_volume(self):
        """Get the volume.

        Integer in range [0..100] or :class:`None` if unknown.

        The volume scale is linear.
        """
        if self._mixer is None:
            return None

        with _mixer_error_handling(self._mixer):
            volume = self._mixer.get_volume().get()
            volume is None or validation.check_integer(volume, min=0, max=100)
            return volume

        return None

    def set_volume(self, volume):
        """Set the volume.

        The volume is defined as an integer in range [0..100].

        The volume scale is linear.

        Returns :class:`True` if call is successful, otherwise :class:`False`.
        """
        validation.check_integer(volume, min=0, max=100)

        if self._mixer is None:
            return False  # TODO: 2.0 return None

        with _mixer_error_handling(self._mixer):
            result = self._mixer.set_volume(volume).get()
            validation.check_instance(result, bool)
            return result

        return False

    def get_mute(self):
        """Get mute state.

        :class:`True` if muted, :class:`False` unmuted, :class:`None` if
        unknown.
        """
        if self._mixer is None:
            return None

        with _mixer_error_handling(self._mixer):
            mute = self._mixer.get_mute().get()
            mute is None or validation.check_instance(mute, bool)
            return mute

        return None

    def set_mute(self, mute):
        """Set mute state.

        :class:`True` to mute, :class:`False` to unmute.

        Returns :class:`True` if call is successful, otherwise :class:`False`.
        """
        validation.check_boolean(mute)
        if self._mixer is None:
            return False  # TODO: 2.0 return None

        with _mixer_error_handling(self._mixer):
            result = self._mixer.set_mute(bool(mute)).get()
            validation.check_instance(result, bool)
            return result

        return False

    def _save_state(self):
        return MixerState(volume=self.get_volume(),
                          mute=self.get_mute())

    def _load_state(self, state, coverage):
        if state and 'mixer' in coverage:
            self.set_mute(state.mute)
            if state.volume:
                self.set_volume(state.volume)



from __future__ import unicode_literals

import os

import mock

import pytest

from mopidy.internal import xdg


@pytest.yield_fixture
def environ():
    patcher = mock.patch.dict(os.environ, clear=True)
    yield patcher.start()
    patcher.stop()


def test_cache_dir_default(environ):
    assert xdg.get_dirs()['XDG_CACHE_DIR'] == os.path.expanduser(b'~/.cache')


def test_cache_dir_from_env(environ):
    os.environ['XDG_CACHE_HOME'] = b'/foo/bar'

    assert xdg.get_dirs()['XDG_CACHE_DIR'] == b'/foo/bar'
    assert type(xdg.get_dirs()['XDG_CACHE_DIR']) == bytes


def test_config_dir_default(environ):
    assert xdg.get_dirs()['XDG_CONFIG_DIR'] == os.path.expanduser(b'~/.config')


def test_config_dir_from_env(environ):
    os.environ['XDG_CONFIG_HOME'] = b'/foo/bar'

    assert xdg.get_dirs()['XDG_CONFIG_DIR'] == b'/foo/bar'
    assert type(xdg.get_dirs()['XDG_CONFIG_DIR']) == bytes


def test_data_dir_default(environ):
    assert xdg.get_dirs()['XDG_DATA_DIR'] == os.path.expanduser(
        b'~/.local/share')


def test_data_dir_from_env(environ):
    os.environ['XDG_DATA_HOME'] = b'/foo/bar'

    assert xdg.get_dirs()['XDG_DATA_DIR'] == b'/foo/bar'
    assert type(xdg.get_dirs()['XDG_DATA_DIR']) == bytes


def test_user_dirs(environ, tmpdir):
    os.environ['XDG_CONFIG_HOME'] = str(tmpdir)

    with open(os.path.join(str(tmpdir), 'user-dirs.dirs'), 'wb') as fh:
        fh.write(b'# Some comments\n')
        fh.write(b'XDG_MUSIC_DIR="$HOME/Music2"\n')

    result = xdg.get_dirs()

    assert result['XDG_MUSIC_DIR'] == os.path.expanduser(b'~/Music2')
    assert type(result['XDG_MUSIC_DIR']) == bytes
    assert 'XDG_DOWNLOAD_DIR' not in result


def test_user_dirs_when_no_dirs_file(environ, tmpdir):
    os.environ['XDG_CONFIG_HOME'] = str(tmpdir)

    result = xdg.get_dirs()

    assert 'XDG_MUSIC_DIR' not in result
    assert 'XDG_DOWNLOAD_DIR' not in result



from __future__ import absolute_import, division, unicode_literals

import datetime
import logging
import re

from mopidy.compat import urllib
from mopidy.mpd import exceptions, protocol, translator

logger = logging.getLogger(__name__)


def _check_playlist_name(name):
    if re.search('[/\n\r]', name):
        raise exceptions.MpdInvalidPlaylistName()


def _get_playlist(context, name, must_exist=True):
    playlist = None
    uri = context.lookup_playlist_uri_from_name(name)
    if uri:
        playlist = context.core.playlists.lookup(uri).get()
    if must_exist and playlist is None:
        raise exceptions.MpdNoExistError('No such playlist')
    return playlist


@protocol.commands.add('listplaylist')
def listplaylist(context, name):
    """
    *musicpd.org, stored playlists section:*

        ``listplaylist {NAME}``

        Lists the files in the playlist ``NAME.m3u``.

    Output format::

        file: relative/path/to/file1.flac
        file: relative/path/to/file2.ogg
        file: relative/path/to/file3.mp3
    """
    playlist = _get_playlist(context, name)
    return ['file: %s' % t.uri for t in playlist.tracks]


@protocol.commands.add('listplaylistinfo')
def listplaylistinfo(context, name):
    """
    *musicpd.org, stored playlists section:*

        ``listplaylistinfo {NAME}``

        Lists songs in the playlist ``NAME.m3u``.

    Output format:

        Standard track listing, with fields: file, Time, Title, Date,
        Album, Artist, Track
    """
    playlist = _get_playlist(context, name)
    track_uris = [track.uri for track in playlist.tracks]
    tracks_map = context.core.library.lookup(uris=track_uris).get()
    tracks = []
    for uri in track_uris:
        tracks.extend(tracks_map[uri])
    playlist = playlist.replace(tracks=tracks)
    return translator.playlist_to_mpd_format(playlist)


@protocol.commands.add('listplaylists')
def listplaylists(context):
    """
    *musicpd.org, stored playlists section:*

        ``listplaylists``

        Prints a list of the playlist directory.

        After each playlist name the server sends its last modification
        time as attribute ``Last-Modified`` in ISO 8601 format. To avoid
        problems due to clock differences between clients and the server,
        clients should not compare this value with their local clock.

    Output format::

        playlist: a
        Last-Modified: 2010-02-06T02:10:25Z
        playlist: b
        Last-Modified: 2010-02-06T02:11:08Z

    *Clarifications:*

    - ncmpcpp 0.5.10 segfaults if we return 'playlist: ' on a line, so we must
      ignore playlists without names, which isn't very useful anyway.
    """
    last_modified = _get_last_modified()
    result = []
    for playlist_ref in context.core.playlists.as_list().get():
        if not playlist_ref.name:
            continue
        name = context.lookup_playlist_name_from_uri(playlist_ref.uri)
        result.append(('playlist', name))
        result.append(('Last-Modified', last_modified))
    return result


# TODO: move to translators?
def _get_last_modified(last_modified=None):
    """Formats last modified timestamp of a playlist for MPD.

    Time in UTC with second precision, formatted in the ISO 8601 format, with
    the "Z" time zone marker for UTC. For example, "1970-01-01T00:00:00Z".
    """
    if last_modified is None:
        # If unknown, assume the playlist is modified
        dt = datetime.datetime.utcnow()
    else:
        dt = datetime.datetime.utcfromtimestamp(last_modified / 1000.0)
    dt = dt.replace(microsecond=0)
    return '%sZ' % dt.isoformat()


@protocol.commands.add('load', playlist_slice=protocol.RANGE)
def load(context, name, playlist_slice=slice(0, None)):
    """
    *musicpd.org, stored playlists section:*

        ``load {NAME} [START:END]``

        Loads the playlist into the current queue. Playlist plugins are
        supported. A range may be specified to load only a part of the
        playlist.

    *Clarifications:*

    - ``load`` appends the given playlist to the current playlist.

    - MPD 0.17.1 does not support open-ended ranges, i.e. without end
      specified, for the ``load`` command, even though MPD's general range docs
      allows open-ended ranges.

    - MPD 0.17.1 does not fail if the specified range is outside the playlist,
      in either or both ends.
    """
    playlist = _get_playlist(context, name)
    track_uris = [track.uri for track in playlist.tracks[playlist_slice]]
    context.core.tracklist.add(uris=track_uris).get()


@protocol.commands.add('playlistadd')
def playlistadd(context, name, track_uri):
    """
    *musicpd.org, stored playlists section:*

        ``playlistadd {NAME} {URI}``

        Adds ``URI`` to the playlist ``NAME.m3u``.

        ``NAME.m3u`` will be created if it does not exist.
    """
    _check_playlist_name(name)
    old_playlist = _get_playlist(context, name, must_exist=False)
    if not old_playlist:
        # Create new playlist with this single track
        lookup_res = context.core.library.lookup(uris=[track_uri]).get()
        tracks = [
            track
            for uri_tracks in lookup_res.values()
            for track in uri_tracks]
        _create_playlist(context, name, tracks)
    else:
        # Add track to existing playlist
        lookup_res = context.core.library.lookup(uris=[track_uri]).get()
        new_tracks = [
            track
            for uri_tracks in lookup_res.values()
            for track in uri_tracks]
        new_playlist = old_playlist.replace(
            tracks=list(old_playlist.tracks) + new_tracks)
        saved_playlist = context.core.playlists.save(new_playlist).get()
        if saved_playlist is None:
            playlist_scheme = urllib.parse.urlparse(old_playlist.uri).scheme
            uri_scheme = urllib.parse.urlparse(track_uri).scheme
            raise exceptions.MpdInvalidTrackForPlaylist(
                playlist_scheme, uri_scheme)


def _create_playlist(context, name, tracks):
    """
    Creates new playlist using backend appropriate for the given tracks
    """
    uri_schemes = {urllib.parse.urlparse(t.uri).scheme for t in tracks}
    for scheme in uri_schemes:
        new_playlist = context.core.playlists.create(name, scheme).get()
        if new_playlist is None:
            logger.debug(
                "Backend for scheme %s can't create playlists", scheme)
            continue  # Backend can't create playlists at all
        new_playlist = new_playlist.replace(tracks=tracks)
        saved_playlist = context.core.playlists.save(new_playlist).get()
        if saved_playlist is not None:
            return  # Created and saved
        else:
            continue  # Failed to save using this backend
    # Can't use backend appropriate for passed URI schemes, use default one
    default_scheme = context.dispatcher.config[
        'mpd']['default_playlist_scheme']
    new_playlist = context.core.playlists.create(name, default_scheme).get()
    if new_playlist is None:
        # If even MPD's default backend can't save playlist, everything is lost
        logger.warning("MPD's default backend can't create playlists")
        raise exceptions.MpdFailedToSavePlaylist(default_scheme)
    new_playlist = new_playlist.replace(tracks=tracks)
    saved_playlist = context.core.playlists.save(new_playlist).get()
    if saved_playlist is None:
        uri_scheme = urllib.parse.urlparse(new_playlist.uri).scheme
        raise exceptions.MpdFailedToSavePlaylist(uri_scheme)


@protocol.commands.add('playlistclear')
def playlistclear(context, name):
    """
    *musicpd.org, stored playlists section:*

        ``playlistclear {NAME}``

        Clears the playlist ``NAME.m3u``.

    The playlist will be created if it does not exist.
    """
    _check_playlist_name(name)
    playlist = _get_playlist(context, name, must_exist=False)
    if not playlist:
        playlist = context.core.playlists.create(name).get()

    # Just replace tracks with empty list and save
    playlist = playlist.replace(tracks=[])
    if context.core.playlists.save(playlist).get() is None:
        raise exceptions.MpdFailedToSavePlaylist(
            urllib.parse.urlparse(playlist.uri).scheme)


@protocol.commands.add('playlistdelete', songpos=protocol.UINT)
def playlistdelete(context, name, songpos):
    """
    *musicpd.org, stored playlists section:*

        ``playlistdelete {NAME} {SONGPOS}``

        Deletes ``SONGPOS`` from the playlist ``NAME.m3u``.
    """
    _check_playlist_name(name)
    playlist = _get_playlist(context, name)

    try:
        # Convert tracks to list and remove requested
        tracks = list(playlist.tracks)
        tracks.pop(songpos)
    except IndexError:
        raise exceptions.MpdArgError('Bad song index')

    # Replace tracks and save playlist
    playlist = playlist.replace(tracks=tracks)
    saved_playlist = context.core.playlists.save(playlist).get()
    if saved_playlist is None:
        raise exceptions.MpdFailedToSavePlaylist(
            urllib.parse.urlparse(playlist.uri).scheme)


@protocol.commands.add(
    'playlistmove', from_pos=protocol.UINT, to_pos=protocol.UINT)
def playlistmove(context, name, from_pos, to_pos):
    """
    *musicpd.org, stored playlists section:*

        ``playlistmove {NAME} {SONGID} {SONGPOS}``

        Moves ``SONGID`` in the playlist ``NAME.m3u`` to the position
        ``SONGPOS``.

    *Clarifications:*

    - The second argument is not a ``SONGID`` as used elsewhere in the protocol
      documentation, but just the ``SONGPOS`` to move *from*, i.e.
      ``playlistmove {NAME} {FROM_SONGPOS} {TO_SONGPOS}``.
    """
    if from_pos == to_pos:
        return

    _check_playlist_name(name)
    playlist = _get_playlist(context, name)
    if from_pos == to_pos:
        return  # Nothing to do

    try:
        # Convert tracks to list and perform move
        tracks = list(playlist.tracks)
        track = tracks.pop(from_pos)
        tracks.insert(to_pos, track)
    except IndexError:
        raise exceptions.MpdArgError('Bad song index')

    # Replace tracks and save playlist
    playlist = playlist.replace(tracks=tracks)
    saved_playlist = context.core.playlists.save(playlist).get()
    if saved_playlist is None:
        raise exceptions.MpdFailedToSavePlaylist(
            urllib.parse.urlparse(playlist.uri).scheme)


@protocol.commands.add('rename')
def rename(context, old_name, new_name):
    """
    *musicpd.org, stored playlists section:*

        ``rename {NAME} {NEW_NAME}``

        Renames the playlist ``NAME.m3u`` to ``NEW_NAME.m3u``.
    """
    _check_playlist_name(old_name)
    _check_playlist_name(new_name)

    old_playlist = _get_playlist(context, old_name)

    if _get_playlist(context, new_name, must_exist=False):
        raise exceptions.MpdExistError('Playlist already exists')
    # TODO: should we purge the mapping in an else?

    # Create copy of the playlist and remove original
    uri_scheme = urllib.parse.urlparse(old_playlist.uri).scheme
    new_playlist = context.core.playlists.create(new_name, uri_scheme).get()
    new_playlist = new_playlist.replace(tracks=old_playlist.tracks)
    saved_playlist = context.core.playlists.save(new_playlist).get()

    if saved_playlist is None:
        raise exceptions.MpdFailedToSavePlaylist(uri_scheme)
    context.core.playlists.delete(old_playlist.uri).get()


@protocol.commands.add('rm')
def rm(context, name):
    """
    *musicpd.org, stored playlists section:*

        ``rm {NAME}``

        Removes the playlist ``NAME.m3u`` from the playlist directory.
    """
    _check_playlist_name(name)
    uri = context.lookup_playlist_uri_from_name(name)
    if not uri:
        raise exceptions.MpdNoExistError('No such playlist')
    context.core.playlists.delete(uri).get()


@protocol.commands.add('save')
def save(context, name):
    """
    *musicpd.org, stored playlists section:*

        ``save {NAME}``

        Saves the current playlist to ``NAME.m3u`` in the playlist
        directory.
    """
    _check_playlist_name(name)
    tracks = context.core.tracklist.get_tracks().get()
    playlist = _get_playlist(context, name, must_exist=False)
    if not playlist:
        # Create new playlist
        _create_playlist(context, name, tracks)
    else:
        # Overwrite existing playlist
        new_playlist = playlist.replace(tracks=tracks)
        saved_playlist = context.core.playlists.save(new_playlist).get()
        if saved_playlist is None:
            raise exceptions.MpdFailedToSavePlaylist(
                urllib.parse.urlparse(playlist.uri).scheme)



from __future__ import unicode_literals

import contextlib
import re
import warnings

from mopidy import compat


# Messages used in deprecation warnings are collected here so we can target
# them easily when ignoring warnings.
_MESSAGES = {
    # Deprecated features mpd:
    'mpd.protocol.playback.pause:state_arg':
        'The use of pause command w/o the PAUSE argument is deprecated.',
    'mpd.protocol.current_playlist.playlist':
        'Do not use this, instead use playlistinfo',

    # Deprecated features in core playback:
    'core.playback.play:tl_track_kwargs':
        'playback.play() with "tl_track" argument is pending deprecation use '
        '"tlid" instead',

    # Deprecated features in core tracklist:
    'core.tracklist.add:tracks_arg':
        'tracklist.add() "tracks" argument is deprecated',

    'core.tracklist.eot_track':
        'tracklist.eot_track() is pending deprecation, use '
        'tracklist.get_eot_tlid()',
    'core.tracklist.next_track':
        'tracklist.next_track() is pending deprecation, use '
        'tracklist.get_next_tlid()',
    'core.tracklist.previous_track':
        'tracklist.previous_track() is pending deprecation, use '
        'tracklist.get_previous_tlid()',
}


def warn(msg_id, pending=False):
    if pending:
        category = PendingDeprecationWarning
    else:
        category = DeprecationWarning
    warnings.warn(_MESSAGES.get(msg_id, msg_id), category)


@contextlib.contextmanager
def ignore(ids=None):
    with warnings.catch_warnings():
        if isinstance(ids, compat.string_types):
            ids = [ids]

        if ids:
            for msg_id in ids:
                msg = re.escape(_MESSAGES.get(msg_id, msg_id))
                warnings.filterwarnings('ignore', msg, DeprecationWarning)
        else:
            warnings.filterwarnings('ignore', category=DeprecationWarning)
        yield



from __future__ import absolute_import, unicode_literals

import gzip
import json
import logging
import os
import tempfile

from mopidy import models
from mopidy.internal import encoding

logger = logging.getLogger(__name__)


def load(path):
    """
    Deserialize data from file.

    :param path: full path to import file
    :type path: bytes
    :return: deserialized data
    :rtype: dict
    """
    # Todo: raise an exception in case of error?
    if not os.path.isfile(path):
        logger.info('File does not exist: %s', path)
        return {}
    try:
        with gzip.open(path, 'rb') as fp:
            return json.load(fp, object_hook=models.model_json_decoder)
    except (IOError, ValueError) as error:
        logger.warning(
            'Loading JSON failed: %s',
            encoding.locale_decode(error))
        return {}


def dump(path, data):
    """
    Serialize data to file.

    :param path: full path to export file
    :type path: bytes
    :param data: dictionary containing data to save
    :type data: dict
    """
    directory, basename = os.path.split(path)

    # TODO: cleanup directory/basename.* files.
    tmp = tempfile.NamedTemporaryFile(
        prefix=basename + '.', dir=directory, delete=False)

    try:
        with gzip.GzipFile(fileobj=tmp, mode='wb') as fp:
            json.dump(data, fp, cls=models.ModelJSONEncoder,
                      indent=2, separators=(',', ': '))
        os.rename(tmp.name, path)
    finally:
        if os.path.exists(tmp.name):
            os.remove(tmp.name)



from __future__ import absolute_import, unicode_literals

import unittest

import mock

import pykka

from mopidy import core
from mopidy.internal import deprecation
from mopidy.mpd import session, uri_mapper

from tests import dummy_audio, dummy_backend, dummy_mixer


class MockConnection(mock.Mock):

    def __init__(self, *args, **kwargs):
        super(MockConnection, self).__init__(*args, **kwargs)
        self.host = mock.sentinel.host
        self.port = mock.sentinel.port
        self.response = []

    def queue_send(self, data):
        lines = (line for line in data.split('\n') if line)
        self.response.extend(lines)


class BaseTestCase(unittest.TestCase):
    enable_mixer = True

    def get_config(self):
        return {
            'core': {
                'max_tracklist_length': 10000
            },
            'mpd': {
                'password': None,
                'default_playlist_scheme': 'dummy',
            }
        }

    def setUp(self):  # noqa: N802
        if self.enable_mixer:
            self.mixer = dummy_mixer.create_proxy()
        else:
            self.mixer = None
        self.audio = dummy_audio.create_proxy()
        self.backend = dummy_backend.create_proxy(audio=self.audio)

        with deprecation.ignore():
            self.core = core.Core.start(
                self.get_config(),
                audio=self.audio,
                mixer=self.mixer,
                backends=[self.backend]).proxy()

        self.uri_map = uri_mapper.MpdUriMapper(self.core)
        self.connection = MockConnection()
        self.session = session.MpdSession(
            self.connection, config=self.get_config(), core=self.core,
            uri_map=self.uri_map)
        self.dispatcher = self.session.dispatcher
        self.context = self.dispatcher.context

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def send_request(self, request):
        self.connection.response = []
        request = '%s\n' % request.encode('utf-8')
        self.session.on_receive({'received': request})
        return self.connection.response

    def assertNoResponse(self):  # noqa: N802
        self.assertEqual([], self.connection.response)

    def assertInResponse(self, value):  # noqa: N802
        self.assertIn(
            value, self.connection.response,
            'Did not find {} in {}'.format(
                repr(value), repr(self.connection.response)))

    def assertOnceInResponse(self, value):  # noqa: N802
        matched = len([r for r in self.connection.response if r == value])
        self.assertEqual(
            1, matched,
            'Expected to find {} once in {}'.format(
                repr(value), repr(self.connection.response)))

    def assertNotInResponse(self, value):  # noqa: N802
        self.assertNotIn(
            value, self.connection.response,
            'Found {} in {}'.format(
                repr(value), repr(self.connection.response)))

    def assertEqualResponse(self, value):  # noqa: N802
        self.assertEqual(1, len(self.connection.response))
        self.assertEqual(value, self.connection.response[0])



from __future__ import absolute_import, unicode_literals

import functools
import logging
import os

import tornado.escape
import tornado.ioloop
import tornado.web
import tornado.websocket

import mopidy
from mopidy import core, models
from mopidy.compat import urllib
from mopidy.internal import encoding, jsonrpc


logger = logging.getLogger(__name__)


def make_mopidy_app_factory(apps, statics):
    def mopidy_app_factory(config, core):
        if not config['http']['csrf_protection']:
            logger.warn(
                'HTTP Cross-Site Request Forgery protection is disabled')
        allowed_origins = {
            x.lower() for x in config['http']['allowed_origins'] if x
        }
        return [
            (r'/ws/?', WebSocketHandler, {
                'core': core,
                'allowed_origins': allowed_origins,
                'csrf_protection': config['http']['csrf_protection'],
            }),
            (r'/rpc', JsonRpcHandler, {
                'core': core,
                'allowed_origins': allowed_origins,
                'csrf_protection': config['http']['csrf_protection'],
            }),
            (r'/(.+)', StaticFileHandler, {
                'path': os.path.join(os.path.dirname(__file__), 'data'),
            }),
            (r'/', ClientListHandler, {
                'apps': apps,
                'statics': statics,
            }),
        ]
    return mopidy_app_factory


def make_jsonrpc_wrapper(core_actor):
    inspector = jsonrpc.JsonRpcInspector(
        objects={
            'core.get_uri_schemes': core.Core.get_uri_schemes,
            'core.get_version': core.Core.get_version,
            'core.history': core.HistoryController,
            'core.library': core.LibraryController,
            'core.mixer': core.MixerController,
            'core.playback': core.PlaybackController,
            'core.playlists': core.PlaylistsController,
            'core.tracklist': core.TracklistController,
        })
    return jsonrpc.JsonRpcWrapper(
        objects={
            'core.describe': inspector.describe,
            'core.get_uri_schemes': core_actor.get_uri_schemes,
            'core.get_version': core_actor.get_version,
            'core.history': core_actor.history,
            'core.library': core_actor.library,
            'core.mixer': core_actor.mixer,
            'core.playback': core_actor.playback,
            'core.playlists': core_actor.playlists,
            'core.tracklist': core_actor.tracklist,
        },
        decoders=[models.model_json_decoder],
        encoders=[models.ModelJSONEncoder]
    )


def _send_broadcast(client, msg):
    # We could check for client.ws_connection, but we don't really
    # care why the broadcast failed, we just want the rest of them
    # to succeed, so catch everything.
    try:
        client.write_message(msg)
    except Exception as e:
        error_msg = encoding.locale_decode(e)
        logger.debug('Broadcast of WebSocket message to %s failed: %s',
                     client.request.remote_ip, error_msg)
        # TODO: should this do the same cleanup as the on_message code?


class WebSocketHandler(tornado.websocket.WebSocketHandler):

    # XXX This set is shared by all WebSocketHandler objects. This isn't
    # optimal, but there's currently no use case for having more than one of
    # these anyway.
    clients = set()

    @classmethod
    def broadcast(cls, msg):
        loop = tornado.ioloop.IOLoop.current()

        # This can be called from outside the Tornado ioloop, so we need to
        # safely cross the thread boundary by adding a callback to the loop.
        for client in cls.clients:
            # One callback per client to keep time we hold up the loop short
            loop.add_callback(functools.partial(_send_broadcast, client, msg))

    def initialize(self, core, allowed_origins, csrf_protection):
        self.jsonrpc = make_jsonrpc_wrapper(core)
        self.allowed_origins = allowed_origins
        self.csrf_protection = csrf_protection

    def open(self):
        self.set_nodelay(True)
        self.clients.add(self)
        logger.debug(
            'New WebSocket connection from %s', self.request.remote_ip)

    def on_close(self):
        self.clients.discard(self)
        logger.debug(
            'Closed WebSocket connection from %s',
            self.request.remote_ip)

    def on_message(self, message):
        if not message:
            return

        logger.debug(
            'Received WebSocket message from %s: %r',
            self.request.remote_ip, message)

        try:
            response = self.jsonrpc.handle_json(
                tornado.escape.native_str(message))
            if response and self.write_message(response):
                logger.debug(
                    'Sent WebSocket message to %s: %r',
                    self.request.remote_ip, response)
        except Exception as e:
            error_msg = encoding.locale_decode(e)
            logger.error('WebSocket request error: %s', error_msg)
            self.close()

    def check_origin(self, origin):
        if not self.csrf_protection:
            return True
        return check_origin(origin, self.request.headers, self.allowed_origins)


def set_mopidy_headers(request_handler):
    request_handler.set_header('Cache-Control', 'no-cache')
    request_handler.set_header(
        'X-Mopidy-Version', mopidy.__version__.encode('utf-8'))


def check_origin(origin, request_headers, allowed_origins):
    if origin is None:
        logger.warn('HTTP request denied for missing Origin header')
        return False
    allowed_origins.add(request_headers.get('Host'))
    parsed_origin = urllib.parse.urlparse(origin).netloc.lower()
    # Some frameworks (e.g. Apache Cordova) use local files. Requests from
    # these files don't really have a sensible Origin so the browser sets the
    # header to something like 'file://' or 'null'. This results here in an
    # empty parsed_origin which we choose to allow.
    if parsed_origin and parsed_origin not in allowed_origins:
        logger.warn('HTTP request denied for Origin "%s"', origin)
        return False
    return True


class JsonRpcHandler(tornado.web.RequestHandler):

    def initialize(self, core, allowed_origins, csrf_protection):
        self.jsonrpc = make_jsonrpc_wrapper(core)
        self.allowed_origins = allowed_origins
        self.csrf_protection = csrf_protection

    def head(self):
        self.set_extra_headers()
        self.finish()

    def post(self):
        if self.csrf_protection:
            content_type = self.request.headers.get('Content-Type', '')
            if content_type != 'application/json':
                self.set_status(415, 'Content-Type must be application/json')
                return

        data = self.request.body
        if not data:
            return

        logger.debug(
            'Received RPC message from %s: %r', self.request.remote_ip, data)

        try:
            self.set_extra_headers()
            response = self.jsonrpc.handle_json(
                tornado.escape.native_str(data))
            if response and self.write(response):
                logger.debug(
                    'Sent RPC message to %s: %r',
                    self.request.remote_ip, response)
        except Exception as e:
            logger.error('HTTP JSON-RPC request error: %s', e)
            self.write_error(500)

    def set_extra_headers(self):
        set_mopidy_headers(self)
        self.set_header('Accept', 'application/json')
        self.set_header('Content-Type', 'application/json; utf-8')

    def options(self):
        if self.csrf_protection:
            origin = self.request.headers.get('Origin')
            if not check_origin(origin, self.request.headers,
                                self.allowed_origins):
                self.set_status(403, 'Access denied for origin %s' % origin)
                return

            self.set_header('Access-Control-Allow-Origin', '%s' % origin)
            self.set_header('Access-Control-Allow-Headers', 'Content-Type')

        self.set_status(204)
        self.finish()


class ClientListHandler(tornado.web.RequestHandler):

    def initialize(self, apps, statics):
        self.apps = apps
        self.statics = statics

    def get_template_path(self):
        return os.path.dirname(__file__)

    def get(self):
        set_mopidy_headers(self)

        names = set()
        for app in self.apps:
            names.add(app['name'])
        for static in self.statics:
            names.add(static['name'])
        names.discard('mopidy')

        self.render('data/clients.html', apps=sorted(list(names)))


class StaticFileHandler(tornado.web.StaticFileHandler):

    def set_extra_headers(self, path):
        set_mopidy_headers(self)


class AddSlashHandler(tornado.web.RequestHandler):

    @tornado.web.addslash
    def prepare(self):
        return super(AddSlashHandler, self).prepare()



from __future__ import absolute_import, unicode_literals

import io

from mopidy.compat import configparser
from mopidy.internal import validation

try:
    import xml.etree.cElementTree as elementtree  # noqa: N813
except ImportError:
    import xml.etree.ElementTree as elementtree  # noqa: N813


def parse(data):
    handlers = {
        detect_extm3u_header: parse_extm3u,
        detect_pls_header: parse_pls,
        detect_asx_header: parse_asx,
        detect_xspf_header: parse_xspf,
    }
    for detector, parser in handlers.items():
        if detector(data):
            return list(parser(data))
    return parse_urilist(data)  # Fallback


def detect_extm3u_header(data):
    return data[0:7].upper() == b'#EXTM3U'


def detect_pls_header(data):
    return data[0:10].lower() == b'[playlist]'


def detect_xspf_header(data):
    data = data[0:150]
    if b'xspf' not in data.lower():
        return False

    try:
        data = io.BytesIO(data)
        for event, element in elementtree.iterparse(data, events=(b'start',)):
            return element.tag.lower() == '{http://xspf.org/ns/0/}playlist'
    except elementtree.ParseError:
        pass
    return False


def detect_asx_header(data):
    data = data[0:50]
    if b'asx' not in data.lower():
        return False

    try:
        data = io.BytesIO(data)
        for event, element in elementtree.iterparse(data, events=(b'start',)):
            return element.tag.lower() == 'asx'
    except elementtree.ParseError:
        pass
    return False


def parse_extm3u(data):
    # TODO: convert non URIs to file URIs.
    found_header = False
    for line in data.splitlines():
        if found_header or line.startswith(b'#EXTM3U'):
            found_header = True
        else:
            continue
        if not line.startswith(b'#') and line.strip():
            yield line.strip()


def parse_pls(data):
    # TODO: convert non URIs to file URIs.
    try:
        cp = configparser.RawConfigParser()
        cp.readfp(io.BytesIO(data))
    except configparser.Error:
        return

    for section in cp.sections():
        if section.lower() != 'playlist':
            continue
        for i in range(cp.getint(section, 'numberofentries')):
            yield cp.get(section, 'file%d' % (i + 1)).strip('\"\'')


def parse_xspf(data):
    try:
        # Last element will be root.
        for event, element in elementtree.iterparse(io.BytesIO(data)):
            element.tag = element.tag.lower()  # normalize
    except elementtree.ParseError:
        return

    ns = 'http://xspf.org/ns/0/'
    path = '{{{}}}tracklist/{{{}}}track'.format(ns, ns)
    for track in element.iterfind(path):
        yield track.findtext('{{{}}}location'.format(ns))


def parse_asx(data):
    try:
        # Last element will be root.
        for event, element in elementtree.iterparse(io.BytesIO(data)):
            element.tag = element.tag.lower()  # normalize
    except elementtree.ParseError:
        return

    for ref in element.findall('entry/ref[@href]'):
        yield ref.get('href', '').strip()

    for entry in element.findall('entry[@href]'):
        yield entry.get('href', '').strip()


def parse_urilist(data):
    result = []
    for line in data.splitlines():
        if not line.strip() or line.startswith(b'#'):
            continue
        try:
            validation.check_uri(line)
        except ValueError:
            return []
        result.append(line)
    return result



# encoding: utf-8

from __future__ import absolute_import, unicode_literals

import unittest

import pytest

from mopidy.internal import playlists


BAD = b'foobarbaz'

EXTM3U = b"""#EXTM3U
#EXTINF:123, Sample artist - Sample title
file:///tmp/foo
#EXTINF:321,Example Artist - Example \xc5\xa7\xc5\x95
file:///tmp/bar

#EXTINF:213,Some Artist - Other title
file:///tmp/baz
"""

URILIST = b"""
file:///tmp/foo
# a comment \xc5\xa7\xc5\x95
file:///tmp/bar

file:///tmp/baz
"""

PLS = b"""[Playlist]
NumberOfEntries=3
File1="file:///tmp/foo"
Title1=Sample Title
Length1=123

File2='file:///tmp/bar'
Title2=Example \xc5\xa7\xc5\x95
Length2=321
File3=file:///tmp/baz
Title3=Other title
Length3=213
Version=2
"""

ASX = b"""<ASX version="3.0">
  <TITLE>Example</TITLE>
  <ENTRY>
    <TITLE>Sample Title</TITLE>
    <REF href="file:///tmp/foo" />
  </ENTRY>
  <ENTRY>
    <TITLE>Example \xc5\xa7\xc5\x95</TITLE>
    <REF href="file:///tmp/bar" />
  </ENTRY>
  <ENTRY>
    <TITLE>Other title</TITLE>
    <REF href="file:///tmp/baz" />
  </ENTRY>
</ASX>
"""

SIMPLE_ASX = b"""<ASX version="3.0">
  <ENTRY href="file:///tmp/foo" />
  <ENTRY href="file:///tmp/bar" />
  <ENTRY href="file:///tmp/baz" />
</ASX>
"""

XSPF = b"""<?xml version="1.0" encoding="UTF-8"?>
<playlist version="1" xmlns="http://xspf.org/ns/0/">
  <trackList>
    <track>
      <title>Sample Title</title>
      <location>file:///tmp/foo</location>
    </track>
    <track>
      <title>Example \xc5\xa7\xc5\x95</title>
      <location>file:///tmp/bar</location>
    </track>
    <track>
      <title>Other title</title>
      <location>file:///tmp/baz</location>
    </track>
  </trackList>
</playlist>
"""

EXPECTED = [b'file:///tmp/foo', b'file:///tmp/bar', b'file:///tmp/baz']


@pytest.mark.parametrize('data,result', [
    (BAD, []),
    (URILIST, EXPECTED),
    (EXTM3U, EXPECTED),
    (PLS, EXPECTED),
    (ASX, EXPECTED),
    (SIMPLE_ASX, EXPECTED),
    (XSPF, EXPECTED),
])
def test_parse(data, result):
    assert playlists.parse(data) == result


class BasePlaylistTest(object):
    valid = None
    invalid = None
    detect = None
    parse = None

    def test_detect_valid_header(self):
        self.assertTrue(self.detect(self.valid))

    def test_detect_invalid_header(self):
        self.assertFalse(self.detect(self.invalid))

    def test_parse_valid_playlist(self):
        uris = list(self.parse(self.valid))
        self.assertEqual(uris, EXPECTED)

    def test_parse_invalid_playlist(self):
        uris = list(self.parse(self.invalid))
        self.assertEqual(uris, [])


class ExtM3uPlaylistTest(BasePlaylistTest, unittest.TestCase):
    valid = EXTM3U
    invalid = BAD
    detect = staticmethod(playlists.detect_extm3u_header)
    parse = staticmethod(playlists.parse_extm3u)


class PlsPlaylistTest(BasePlaylistTest, unittest.TestCase):
    valid = PLS
    invalid = BAD
    detect = staticmethod(playlists.detect_pls_header)
    parse = staticmethod(playlists.parse_pls)


class AsxPlsPlaylistTest(BasePlaylistTest, unittest.TestCase):
    valid = ASX
    invalid = BAD
    detect = staticmethod(playlists.detect_asx_header)
    parse = staticmethod(playlists.parse_asx)


class SimpleAsxPlsPlaylistTest(BasePlaylistTest, unittest.TestCase):
    valid = SIMPLE_ASX
    invalid = BAD
    detect = staticmethod(playlists.detect_asx_header)
    parse = staticmethod(playlists.parse_asx)


class XspfPlaylistTest(BasePlaylistTest, unittest.TestCase):
    valid = XSPF
    invalid = BAD
    detect = staticmethod(playlists.detect_xspf_header)
    parse = staticmethod(playlists.parse_xspf)


class UriListPlaylistTest(unittest.TestCase):
    valid = URILIST
    invalid = BAD
    parse = staticmethod(playlists.parse_urilist)

    def test_parse_valid_playlist(self):
        uris = list(self.parse(self.valid))
        self.assertEqual(uris, EXPECTED)

    def test_parse_invalid_playlist(self):
        uris = list(self.parse(self.invalid))
        self.assertEqual(uris, [])



from __future__ import absolute_import, unicode_literals

import pytest

from mopidy.audio import utils
from mopidy.internal.gi import Gst


class TestCreateBuffer(object):

    def test_creates_buffer(self):
        buf = utils.create_buffer(b'123', timestamp=0, duration=1000000)

        assert isinstance(buf, Gst.Buffer)
        assert buf.pts == 0
        assert buf.duration == 1000000
        assert buf.get_size() == len(b'123')

    def test_fails_if_data_has_zero_length(self):
        with pytest.raises(ValueError) as excinfo:
            utils.create_buffer(b'', timestamp=0, duration=1000000)

        assert 'Cannot create buffer without data' in str(excinfo.value)
